"""Xml_digest module."""

from pineboolib.core.utils import logging

import os
import xmlsig  # type: ignore[import] # noqa: F821
from lxml import etree  # type: ignore[import] # noqa: F821
from cryptography.hazmat import backends  # type: ignore[import] # noqa: F821
from cryptography.hazmat.primitives.serialization import pkcs12  # type: ignore[import] # noqa: F821
from xades import policy, utils, template, XAdESContext  # type: ignore[import] # noqa: F821


from typing import List, Optional, Any, Union

LOGGER = logging.get_logger(__name__)


class XmlDigest:
    """XmlDigest class."""

    _root: "etree._Element"
    _pass: str
    _policy_list: List[str]
    _cert_path: str
    _certificate: Any
    _policy: Optional["policy.GenericPolicyId"]
    _signature: Optional["etree._Element"]
    _is_signed: bool
    _sha: Optional[int]
    _rsa: Optional[int]
    _use_algorithm: str

    def __init__(
        self, file_path_or_xml: Union[str, "etree._Element"], cert_path: str, pwsd_: str = ""
    ) -> None:
        """Initialize."""

        for path in [file_path_or_xml, cert_path]:
            if isinstance(path, str):
                if not os.path.exists(path):
                    raise Exception("%s doesn't exists!" % path)

        self._root = (
            etree.parse(file_path_or_xml).getroot()
            if isinstance(file_path_or_xml, str)  # type: ignore [unreachable]
            else file_path_or_xml
        )
        self._cert_path = cert_path
        self._pass = pwsd_
        self._policy_list = [
            "http://www.facturae.gob.es/politica_de_firma_formato_facturae/politica_de_firma_formato_facturae_v3_1.pdf",
            "Politica de Firma FacturaE v3.1",
            "xmlsig.constants.TransformSha1",
        ]
        self._certificate = None
        self._policy = None
        self._signature = None
        self._is_signed = False
        self._use_algorithm = "sha256"
        self._sha = None
        self._rsa = None

    def set_password(self, pwds_: str = "") -> None:
        """Set password."""

        self._pass = pwds_

    def _load_certificate(self) -> bool:
        """Load certificate."""

        if not self._pass:
            LOGGER.warning("Password is empty!")
            return False
        try:
            with open(self._cert_path, "rb") as cert_file:
                self._certificate = tuple(
                    pkcs12.load_key_and_certificates(
                        cert_file.read(), self._pass.encode(), backends.default_backend()
                    )
                )
        except Exception as error:
            LOGGER.warning("Error loading certificate: %s", str(error))
            return False

        return True

    def _load_policy(self) -> bool:
        """Load policy."""

        if not self._policy_list:
            LOGGER.warning("Policy is empty!")
            return False

        custom_policy: List[Any] = list(self._policy_list)
        if len(custom_policy) == 1:
            custom_policy += ["Policy description"]
        custom_policy.append(self._sha)

        try:
            self._policy = policy.GenericPolicyId(*custom_policy)
        except Exception as error:
            LOGGER.warning("Error loading policies: %s", str(error))
            return False

        return True

    def _load_signature(self) -> bool:
        """Load signature."""

        try:
            unique_id = utils.get_unique_id()
            signature_id = "Signature-%s" % unique_id
            qualifing_id = "QualifyingProperties-%s" % unique_id
            reference_id = "Reference-%s" % unique_id

            self._signature = xmlsig.template.create(
                xmlsig.constants.TransformInclC14N, self._rsa, signature_id
            )

            reference = xmlsig.template.add_reference(
                self._signature, self._sha, uri="", name="Reference-%s" % reference_id
            )
            xmlsig.template.add_transform(reference, xmlsig.constants.TransformEnveloped)

            xmlsig.template.add_reference(
                self._signature,
                self._sha,
                uri="#KeyInfoId-%s" % signature_id,
                name="ReferenceKeyInfo",
            )
            xmlsig.template.add_reference(
                self._signature,
                self._sha,
                uri="#SignedProperties-%s" % signature_id,
                uri_type="http://uri.etsi.org/01903#SignedProperties",
            )
            key_info = xmlsig.template.ensure_key_info(
                self._signature, name="KeyInfoId-%s" % signature_id
            )

            data = xmlsig.template.add_x509_data(key_info)

            xmlsig.template.x509_data_add_certificate(data)

            # serial = xmlsig.template.x509_data_add_issuer_serial(data)
            # xmlsig.template.x509_issuer_serial_add_issuer_name(serial)
            # xmlsig.template.x509_issuer_serial_add_serial_number(serial)

            xmlsig.template.add_key_value(key_info)

            qualifying = template.create_qualifying_properties(
                self._signature, name=qualifing_id, etsi="xades"
            )
            utils.ensure_id(qualifying)

            props = template.create_signed_properties(
                qualifying, name="SignedProperties-%s" % signature_id
            )

            signed_do = template.ensure_signed_data_object_properties(props)

            template.add_data_object_format(
                signed_do,
                "#%s" % reference_id,
                description="",
                mime_type="application/xml",
                encoding="UTF-8",
            )

        except Exception as error:
            LOGGER.warning("Error loading signature: %s", str(error))
            return False

        return True

    def set_policy(self, policy_list: List[str]) -> None:
        """Set policy."""

        self._policy_list = policy_list

    def set_algorithm(self, name: str = "sha1") -> None:
        """Set algorithm type."""

        self._use_algorithm = name.lower()

    def sign(self) -> bool:
        """Sign data."""

        self._sha = None
        self._rsa = None

        if self._use_algorithm == "sha1":
            self._sha = xmlsig.constants.TransformSha1
            self._rsa = xmlsig.constants.TransformRsaSha1
        elif self._use_algorithm == "sha256":
            self._sha = xmlsig.constants.TransformSha256
            self._rsa = xmlsig.constants.TransformRsaSha256
        elif self._use_algorithm == "sha512":
            self._sha = xmlsig.constants.TransformSha512
            self._rsa = xmlsig.constants.TransformRsaSha512
        else:
            LOGGER.warning("UNKNOWN algorithm %s", self._use_algorithm)
            return False

        if not self._load_certificate():
            LOGGER.warning("certificate not loaded!")
            return False

        if not self._load_policy():
            LOGGER.warning("policy not loaded!")
            return False

        if not self._load_signature():
            LOGGER.warning("signature not loaded!")
            return False

        try:
            if self._signature is None:
                raise Exception("Signature is empty!")

            self._root.append(self._signature)  # type: ignore [attr-defined]

            if self._policy is None:
                raise Exception("Policy is empty!")
            context = XAdESContext(self._policy)
            if self._certificate is None:
                raise Exception("Certificate is empty!")
            context.load_pkcs12(self._certificate)

            LOGGER.warning("Starting signing")
            # LOGGER.info(
            #    "Policy : %s --> %s --> %s"
            #    % (self._policy, self._policy.identifier, context.policies)
            # )

            # LOGGER.info("Certificate : %s" % str(self._certificate))
            # LOGGER.info("Signature : %s" % self._signature)

            context.sign(self._signature)
            LOGGER.warning("Signing finished sucefully!")
        except Exception as error:
            LOGGER.warning("Error signing: %s" % str(error))
            return False

        self._is_signed = True
        return True

    def signature_value(self) -> str:
        """Return SisgnatureValue field value."""

        if not self._is_signed:
            LOGGER.warning("xml is not signed yet")
            return ""

        for child in self._root:  # type: ignore [attr-defined]
            if str(child.get("Id")).startswith("Signature"):
                for child_elem in child:
                    if "SignatureValue" in child_elem.tag:
                        return child_elem.text or ""

        LOGGER.warning("SignatureValue not found!")
        return ""

    def save_file(self, file_path: str) -> bool:
        """Save signed xml into a file."""

        if not file_path:
            LOGGER.warning("file_path is empty!")
        elif not self._is_signed:
            LOGGER.warning("the xml is not signed")
        else:
            try:
                if os.path.exists(file_path):
                    os.remove(file_path)

                element_tree = etree.ElementTree(self._root)
                element_tree.write(
                    file_path, pretty_print=False, xml_declaration=True, encoding="UTF-8"
                )
                return True
            except Exception as error:
                LOGGER.warning("Error saving file %s: %s", file_path, str(error))

        return False



"""
XMLAction module.
"""


from pineboolib.core.utils import struct, utils_base
from pineboolib.core import garbage_collector

from pineboolib.application.database import pnsqlcursor
from pineboolib import logging, application

from pineboolib.application import load_script

import threading
from typing import Optional, Union, Dict, List, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application import moduleactions  # noqa : F401 # pragma: no cover
    from pineboolib.interfaces import isqlcursor  # noqa: F401 # pragma: no cover
    from xml.etree import ElementTree as ET  # noqa: F401 # pragma: no cover
    from pineboolib.qsa import formdbwidget  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class XMLAction(struct.ActionStruct):
    """
    Information related to actions specified in XML modules.
    """

    _mod: Optional["moduleactions.ModuleActions"]

    __master_widget: Dict[int, "formdbwidget.FormDBWidget"]
    __record_widget: Dict[int, "formdbwidget.FormDBWidget"]

    __cursor: Dict[int, "isqlcursor.ISqlCursor"]

    def __init__(
        self, module: "moduleactions.ModuleActions", action_xml_or_str: Union["ET.Element", str]
    ) -> None:
        """
        Initialize.
        """
        if isinstance(action_xml_or_str, str):
            super().__init__()
            self._name = action_xml_or_str
        else:
            super().__init__(action_xml_or_str)
            self._name = self._v("name") or ""

        self._mod = module
        self._alias = self._v("alias") or ""
        self._description = self._v("description") or ""
        self._caption = self._v("caption") or self._description
        self._master_form = self._v("form") or ""
        self._master_script = self._v("scriptform") or ""
        self._record_form = self._v("formrecord") or ""
        self._record_script = self._v("scriptformrecord") or ""
        self._table = self._v("table") or ""
        self._class_script = self._v("class") or ""
        self._class_orm = self._v("model") or ""

        self.__master_widget = {}
        self.__record_widget = {}

        self.__cursor = {}

    def setCursor(self, cursor: Optional["isqlcursor.ISqlCursor"] = None):
        """Set xmlAction cursor."""
        # LOGGER.warning(
        #    "Seteando cursor para %s %s", self._name, self._master_widget, stack_info=True
        # )
        id_thread = threading.current_thread().ident
        if cursor is not self._cursor:
            del self.__cursor[id_thread]  # type: ignore [index, arg-type] # noqa: F821
            self.__cursor[id_thread] = cursor  # type: ignore [assignment, index] # noqa: F821

    def cursor(self) -> Optional["isqlcursor.ISqlCursor"]:
        """Return xmlAction cursor."""
        if not self._cursor and self._table:
            # LOGGER.warning("Creando cursor para %s %s", self._name, self._master_widget)
            self._cursor = pnsqlcursor.PNSqlCursor(self._name)

        return self._cursor

    def load_widget(self, script_name: str) -> "formdbwidget.FormDBWidget":
        """Return widget."""

        return load_script.load_script(script_name, self)

    def clear_widget(self, widget: Optional["formdbwidget.FormDBWidget"]) -> None:
        """Clear old widget."""

        if widget is not None:
            id_thread = threading.current_thread().ident or 0

            widget.clear_connections()
            proxy_parent = getattr(widget, "_my_proxy", None)
            if proxy_parent is not None and id_thread in proxy_parent.loaded_obj.keys():
                del proxy_parent.loaded_obj[id_thread]

            if widget is self._record_widget:
                del self.__record_widget[id_thread]
            elif widget is self._master_widget:
                del self.__master_widget[id_thread]

            # from PyQt5 import QtCore

            if widget._form is not None:
                # if application.PROJECT.main_window:
                #    # if application.PROJECT.main_window.main_widget is widget._form:
                #    for child in application.PROJECT.main_window.findChildren(QtCore.QObject):
                #        if child is widget._form:
                #            del child

                # obj_form = widget._form
                # Cerrando hijos ...

                # for child in widget._form.findChildren(QtCore.QObject):
                #    if hasattr(child, "_loaded"):
                #        child._top_widget = None
                #    if hasattr(child, "fltable_iface"):
                #        child.fltable_iface = None
                self.clear_form(widget)
                # print(widget._form)

                # garbage_collector.check_delete(obj_form, "widget.form")

            if hasattr(widget, "iface"):
                if hasattr(widget.iface, "ctx"):
                    obj_ctx = widget.iface.ctx  # type: ignore [attr-defined] # noqa: F821
                    widget.iface.ctx.deleteLater()  # type: ignore [attr-defined] # noqa: F821
                    widget.iface.ctx = None  # type: ignore [attr-defined] # noqa: F821
                    garbage_collector.check_delete(obj_ctx, "widget.iface.ctx")
                obj_iface = widget.iface
                del widget.iface
                widget.iface = None
                garbage_collector.check_delete(obj_iface, "widget.iface")

            garbage_collector.check_delete(widget, "widget")
            # del widget

    def clear_form(self, widget: "formdbwidget.FormDBWidget") -> bool:
        """Delete form associated ."""
        if widget._form is not None:
            widget._form.setParent(None)  # type: ignore [call-overload] # noqa: F821
            widget._form.deleteLater()
            widget._form = None

        return True

    def is_form_loaded(self, widget: Optional["formdbwidget.FormDBWidget"]) -> bool:
        """Return if widget.form is loaded."""

        form = getattr(widget, "_form", None)
        return getattr(form, "_loaded", False)

    def load_class(self):
        """Load and return class."""
        if self._class_script:
            return load_script.load_class(self._class_script)

    def load_master_widget(self) -> "formdbwidget.FormDBWidget":
        """
        Load master form.
        """
        if not self.is_form_loaded(self._master_widget):
            if self._table or (not self._table and self._master_widget is None):
                self._master_widget = self.load_widget(
                    self._master_script if self._table else self._name
                )

        if self._master_widget is None:
            raise Exception("After load_master_widget, no widget was loaded")

        return self._master_widget

    def load_record_widget(self) -> "formdbwidget.FormDBWidget":
        """
        Load FLFormRecordDB by default.

        """
        if self._record_widget is None:
            self._record_widget = self.load_widget(self._record_script)
        elif self._record_widget._form is not None and not self.is_form_loaded(self._record_widget):
            self.clear_widget(self._record_widget)

            self._record_widget = self.load_widget(self._record_script)

        if self._record_widget is None:
            raise Exception("After load_record_widget, no widget was loaded")

        return self._record_widget

    def load_master_form(self) -> None:
        """Load master_widget.form."""

        if not self.is_form_loaded(self._master_widget):
            if self._master_widget is None:
                self._master_widget = self.load_master_widget()

            form = None
            if not utils_base.is_library():
                form = application.PROJECT.conn_manager.managerModules().createForm(action=self)
                if form is None:
                    raise Exception("After createForm, no form was loaded")
                else:
                    form._loaded = True

            self._master_widget._form = form  # type: ignore [assignment] # noqa: F821

            if form is not None:
                preload_main_filter = getattr(form.iface, "preloadMainFilter", None)
                if preload_main_filter is not None:
                    value = preload_main_filter()
                    if value is not None and form.cursor_:
                        form.cursor_.setMainFilter(value, False)

    def load_record_form(self, cursor: Optional["isqlcursor.ISqlCursor"] = None) -> None:
        """Load record_widget.form."""

        if not self.is_form_loaded(self._record_widget):
            if self._record_widget is None:
                self._record_widget = self.load_record_widget()

            form = None
            if not utils_base.is_library():
                form = application.PROJECT.conn_manager.managerModules().createFormRecord(
                    action=self, parent_or_cursor=cursor
                )

                if form is None:
                    raise Exception("After createFormRecord, no form was loaded")
                else:
                    form._loaded = True

            if cursor is not None:
                if form is not None:
                    form.setCursor(cursor)
                else:
                    LOGGER.warning(
                        "add cursor?. the form does not exist!!. action: %s record_form: %s",
                        self._name,
                        self._record_form,
                    )

            self.clear_form(self._record_widget)
            self._record_widget._form = form  # type: ignore [assignment] # noqa: F821

    def openDefaultForm(self) -> None:
        """
        Open Main FLForm specified on defaults.
        """
        LOGGER.info("Opening default form for Action %s", self._name)
        self.load_master_form()
        if self._master_widget is not None and self._master_widget.form is not None:
            self._master_widget.form.show()

    def openDefaultFormRecord(self, cursor: "isqlcursor.ISqlCursor", wait: bool = True) -> None:
        """
        Open FLFormRecord specified on defaults.

        @param cursor. Cursor a usar por el FLFormRecordDB
        """

        if self.is_form_loaded(self._record_widget):
            if self._record_widget is not None and self._record_widget.form is not None:
                if self._record_widget.form._showed:
                    msg = "Ya hay abierto un formulario de edición de registro para esta tabla.\n"
                    "No se abrirán mas para evitar ciclos repetitivos de edición de registros."
                    application.PROJECT.message_manager().send(
                        "msgBoxInfo", None, [msg, None, "Aviso"]
                    )

            LOGGER.warning("formRecord%s is already loaded!", self._record_form)
            return

        self.load_record_form(cursor)

        if self._record_widget is not None and self._record_widget.form is not None:
            if wait:
                self._record_widget.form.show_and_wait()
            else:
                self._record_widget.form.show()

    def formRecordWidget(self) -> "formdbwidget.FormDBWidget":
        """
        Return formrecord widget.

        This is needed because sometimes there isn't a FLFormRecordDB initialized yet.
        @return wigdet del formRecord.
        """

        return self.load_record_widget()

    def execMainScript(self, action_name: str) -> None:
        """
        Execute function for main action.
        """
        application.PROJECT.call("form%s.main" % action_name, [], None, True)

    def execDefaultScript(self):
        """
        Execute script specified on default.
        """
        widget = self.load_master_widget()

        base_function = getattr(widget, "iface", widget)

        main = getattr(base_function, "main", None)
        if main is None:
            raise Exception("main function not found!")
        else:
            main()

    def unknownSlot(self) -> None:
        """Log error for actions with unknown slots or scripts."""

        LOGGER.error("Executing unknown script for Action %s", self._name)

    def get_master_widget(self) -> "formdbwidget.FormDBWidget":
        """Return master widget."""

        self.thread_cleaner()

        id_thread = threading.current_thread().ident
        if id_thread not in self.__master_widget.keys():
            self.__master_widget[id_thread] = None  # type: ignore [assignment, index] # noqa: F821

        return self.__master_widget[id_thread]  # type: ignore [index] # noqa: F821

    def set_master_widget(self, widget: Optional["formdbwidget.FormDBWidget"] = None) -> None:
        """Set master widget."""

        id_thread = threading.current_thread().ident
        self.__master_widget[id_thread] = widget  # type: ignore [assignment, index] # noqa: F821

    def get_record_widget(self) -> "formdbwidget.FormDBWidget":
        """Return record widget."""

        self.thread_cleaner()

        id_thread = threading.current_thread().ident
        if id_thread not in self.__record_widget.keys():
            self.__record_widget[id_thread] = None  # type: ignore [assignment, index] # noqa: F821

        return self.__record_widget[id_thread]  # type: ignore [index] # noqa: F821

    def set_record_widget(self, widget: Optional["formdbwidget.FormDBWidget"] = None) -> None:
        """Set record widget."""

        id_thread = threading.current_thread().ident
        self.__record_widget[id_thread] = widget  # type: ignore [assignment, index] # noqa: F821

    def get_action_cursor(self) -> Optional["isqlcursor.ISqlCursor"]:
        """Return action cursor."""

        self.thread_cleaner()

        id_thread = threading.current_thread().ident
        if id_thread not in self.__cursor.keys():
            return None

        return self.__cursor[id_thread]  # type: ignore [index] # noqa: F821

    def set_action_cursor(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Set action cursor."""

        id_thread = threading.current_thread().ident
        self.__cursor[id_thread] = cursor  # type: ignore [index] # noqa: F821

    def thread_cleaner(self) -> None:
        """Clean unused threads."""

        # limpieza
        threads_ids: List[Optional[int]] = [thread.ident for thread in threading.enumerate()]
        obj_list: List[Dict[int, Any]] = [self.__cursor, self.__master_widget, self.__record_widget]
        for obj_ in obj_list:
            for id_thread in list(obj_.keys()):  # type: ignore [attr-defined] # noqa: F821
                if obj_[id_thread] is not None:
                    if id_thread not in threads_ids:
                        obj_to_delete = obj_[id_thread]
                        obj_[id_thread] = None
                        del obj_[id_thread]
                        garbage_collector.check_delete(obj_to_delete, self._name)

    _master_widget = property(get_master_widget, set_master_widget)
    _record_widget = property(get_record_widget, set_record_widget)
    _cursor = property(get_action_cursor, set_action_cursor)



"""Service module."""

import os
import sys

from typing import List


def main() -> None:
    """Service manager."""

    if os.geteuid() != 0:
        exit(
            "You need to have root privileges to run this script.\nPlease try again, this time using 'sudo'. Exiting."
        )

    option = "--help"
    if len(sys.argv) > 1:
        option = sys.argv[1]

    if option in ["-h", "--help"]:
        print(
            "options:\n\t --install : Install service.\n\t --remove  : Remove service.\n\t -h --help : This help."
        )
    elif option in ["--install"]:
        path_initd = _resolve_path()
        if _remove_service(path_initd):
            print("Ok" if _install_service(path_initd) else "Fail")

    elif option in ["--remove"]:
        path_initd = _resolve_path()
        print("Ok" if _remove_service(path_initd) else "Fail")


def _install_service(path_initd: str) -> bool:
    """Remove neew service file."""

    pineboo_service_file_name = os.path.join(path_initd, "pineboo_service")
    try:
        file_ = open(pineboo_service_file_name, "w", encoding="UTF-8")
        file_.writelines(["%s\n" % linea for linea in _script_lines()])
        file_.close()
    except Exception as error:
        print("Error writing file %s : %s" % (pineboo_service_file_name, error))
        return False

    return True


def _remove_service(path_initd: str) -> bool:
    """Remove old service file."""

    pineboo_service_file_name = os.path.join(path_initd, "pineboo_service")
    print("Checking for %s" % pineboo_service_file_name)
    if os.path.exists(pineboo_service_file_name):
        print("Removing older file %s" % pineboo_service_file_name)
        try:
            os.remove(pineboo_service_file_name)
        except Exception as error:
            print("Error deleting file %s : %s" % (pineboo_service_file_name, error))
            return False

    return True


def _resolve_path() -> str:
    """Resolve initd path."""

    path_initd = sys.argv[2] if len(sys.argv) == 3 else None
    if not path_initd:
        for test_path in ["/etc/init.d"]:
            if os.path.exists(test_path):
                path_initd = test_path
                break

    if path_initd:
        if not os.path.exists(path_initd):
            exit("Invalid initd path (%s)" % path_initd)
    else:
        exit("Initd path not found")

    return path_initd


def _script_lines() -> List[str]:
    """Return script lines."""
    from pineboolib.application import PINEBOO_VER

    data: List[str] = []
    data.append("#!/usr/bin/python")
    data.append("")
    data.append("Build by Pineboo %s" % PINEBOO_VER)
    data.append("")
    data.append("from daemon import runner")
    data.append("from pineboolib import application as pineboolib_app")
    data.append("from pineboolib.application.parsers import parser_qsa as qsaparser")
    data.append("from pineboolib.core.settings import CONFIG")
    data.append("from pineboolib.loader.projectconfig import ProjectConfig")
    data.append("")
    data.append("temp_dir=/opt/pineboo_cache")
    data.append("prj_name=None # Path to file project (.xml)")
    data.append(
        "SQL_CONN=None # ProjectConfig(database=_database_, host=_host_,"
        + " port=_port_, type='PostgreSQL (PSYCOPG2)', username=_user_name_, password=_pass_)"
    )
    data.append("")
    data.append("")
    data.append("class App():")
    data.append("    def __init__(self):")
    data.append("        self.stdin_path = '/dev/null'")
    data.append("        self.stdout_path = '/dev/tty'")
    data.append("        self.stderr_path = '/dev/tty'")
    data.append("        self.pidfile_path = '/tmp/pineboo_service.pid'")
    data.append("        # self.pidfile_timeout = 5")
    data.append("    def run(self):")
    data.append("")
    data.append("        qsaparser.USE_THREADS = False")
    data.append("        prj = pineboolib_app.PROJECT")
    data.append("        prj.tmpdir = temp_dir")
    data.append("        prj.conn_manager.REMOVE_CONNECTIONS_AFTER_ATOMIC = True")
    data.append("        prj.conn_manager.set_max_connections_limit(100)")
    data.append("        prj.conn_manager.set_safe_mode(2)")
    data.append("        if prj_name:")
    data.append("            SQL_CONN=ProjectConfig(load_xml=prj_name)")
    data.append("")
    data.append("        if SQL_CONN:")
    data.append("            main.startup_framework(SQL_CONN)")
    data.append("")
    data.append("        prj.no_python_cache = False")
    data.append("        pineboolib_app.SHOW_CURSOR_EVENTS = False")
    data.append("")
    data.append("app = App()")
    data.append("daemon_runner = runner.DaemonRunner(app)")
    data.append("daemon_runner.do_action()")

    return data



"""Test_sysbasetype module."""

import unittest

from pineboolib import application

from pineboolib.loader.main import init_testing, finish_testing


class TestMtdParserGeneral(unittest.TestCase):
    """TestMtdParserGeneral Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""

        init_testing()

    def test_basic_1(self) -> None:
        """Test ORM parser."""

        from pineboolib.application.parsers.parser_mtd import pnmtdparser, pnormmodelsfactory
        import os

        for mtd_name in application.PROJECT.files.keys():
            if mtd_name.endswith(".mtd"):
                file_path = pnmtdparser.mtd_parse(
                    mtd_name, application.PROJECT.files[mtd_name].path()
                )
                if file_path:
                    self.assertTrue(os.path.exists(file_path))
                else:
                    self.assertTrue(False)

        pnormmodelsfactory.load_models()

    def test_basic_3(self) -> None:
        """Test create table."""

        from sqlalchemy import MetaData  # type: ignore

        meta = MetaData()
        meta.create_all(application.PROJECT.conn_manager.mainConn().engine())

    def test_basic_4(self) -> None:
        """Test load model."""
        from pineboolib.qsa import qsa

        flmodules_orm = qsa.orm_("flmodules")
        self.assertTrue(flmodules_orm)

    def test_basic_5(self) -> None:
        """Test regexp."""
        from pineboolib.qsa import qsa

        test_orm = qsa.orm_("fltest5")
        self.assertTrue(test_orm)
        field = test_orm.table_metadata().field("version")
        self.assertTrue(field)
        self.assertEqual(field.regExpValidator(), r"^(\d{1,2}(,\d{1,2})*)?$")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flseqs_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Flseqs(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flseqs class."""

    __tablename__ = "flseqs"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flseqs",
        "alias": "Secuencias",
        "fields": [
            {
                "name": "tabla",
                "alias": "Nombre de la tabla",
                "pk": True,
                "type": "string",
                "length": 255,
                "null": False,
            },
            {
                "name": "campo",
                "alias": "Nombre del campo",
                "ck": True,
                "type": "string",
                "length": 255,
                "null": False,
            },
            {"name": "seq", "alias": "Secuencia", "type": "uint", "null": False},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    tabla = sqlalchemy.Column("tabla", sqlalchemy.String(255), primary_key=True)
    campo = sqlalchemy.Column("campo", sqlalchemy.String(255))
    seq = sqlalchemy.Column("seq", sqlalchemy.BigInteger)


# <--- Fields ---



"""
To detect if we are in mobile mode.
"""

from PyQt6 import QtCore  # type: ignore[import]

MOBILE_MODE = None


def is_mobile_mode() -> bool:
    """
    Return if you are working in mobile mode.

    @return True or False
    """

    global MOBILE_MODE
    if MOBILE_MODE is None:
        MOBILE_MODE = check_mobile_mode()
    return MOBILE_MODE


def check_mobile_mode() -> bool:
    """
    Return if you are working in mobile mode, searching local settings or check QtCore.QSysInfo().productType().

    @return True or False.
    """
    from pineboolib.core import settings

    return (
        True
        if QtCore.QSysInfo().productType() in ("android", "ios")
        else settings.CONFIG.value("ebcomportamiento/mobileMode", False)
    )


# def grand_storage_access():
#    """Grand storage permissions."""

#    sys_info = QtCore.QSysInfo()
#    product_type = sys_info.productType()

#    if product_type == "android":
#        import jnius
# from PyQt6 import QtAndroidExtras  # type: ignore #noqa: F821
# import sys

# android_jni = QtAndroidExtras.QAndroidJniObject("android/content/Context")
# get_system_service = android_jni.callStaticMethod(
#    "android/content/Context", "getSystemService"
# )
# result = QtAndroidExtras.QAndroidJniObject.callStaticMethod(
#    "android/content/Context", "getSystemService"
# )

# else:
#    if not android_jni.callStaticMethod("getSystemService"):
#        sys.exit(32)
# storage_service = get_system_service("android.context.Context.STORAGE_SERVICE")
# storage_service = get_system_service("STORAGE_SERVICE")
# get_system_service = android_jni.callStaticMethod("Context", "getSystemService")
# start_activity = android_jni.callStaticMethod("Context", "startActivity")
# storage_service = get_system_service("STORAGE_SERVICE")

# storage_list = storage_service.getStorageVolumes()
# for storage in storage_list:
#    intent = storage.createAccessIntent()
#    start_activity(intent)



"""Test_flsqlite module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.plugins.sql import flsqlite


class TestFLSqlite(unittest.TestCase):
    """TestFLSqlite Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Basics test 1."""

        driver = flsqlite.FLSQLITE()

        self.assertEqual(driver.formatValueLike("bool", "true", False), "=0")
        self.assertEqual(driver.formatValueLike("date", "27-01-2020", True), "LIKE '%%2020-01-27'")

        self.assertEqual(driver.formatValue("bool", "false", True), "0")
        self.assertEqual(driver.formatValue("time", "", True), "")

        # self.assertFalse(driver.transaction_rollback())
        # self.assertFalse(driver.save_point(0))
        # self.assertFalse(driver.transaction_commit())
        # self.assertFalse(driver.transaction())
        # self.assertFalse(driver.save_point_release(0))

        self.assertEqual(driver.setType("String", 20), "VARCHAR(20)")
        self.assertEqual(driver.setType("sTring", 0), "VARCHAR")
        self.assertEqual(driver.setType("Double"), "FLOAT")
        self.assertEqual(driver.setType("Bool"), "BOOLEAN")
        self.assertEqual(driver.setType("DATE"), "VARCHAR(20)")
        self.assertEqual(driver.setType("pixmap"), "TEXT")
        self.assertEqual(driver.setType("bytearray"), "CLOB")
        self.assertEqual(driver.setType("timestamp"), "DATETIME")
        # self.assertEqual(
        #    driver.process_booleans("'true' AND false AND 'false'"), "1 AND false AND 0"
        # )

    def test_basic_2(self) -> None:
        """Basics test 2."""
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        sql = (
            "CREATE TABLE fltest (id INTEGER PRIMARY KEY,string_field VARCHAR NULL,date_field VARCHAR(20)"
            + " NULL,time_field VARCHAR(20) NULL,double_field FLOAT NULL,bool_field BOOLEAN NULL"
            + ",uint_field INTEGER NULL,bloqueo BOOLEAN NOT NULL,empty_relation VARCHAR(15) NULL,"
            + "int_field INTEGER NULL);CREATE INDEX fltest_pkey ON fltest (id)"
        )
        sql2 = (
            "CREATE TABLE fltest (id INTEGER PRIMARY KEY,string_field VARCHAR NULL,date_field VARCHAR(20)"
            + " NULL,time_field VARCHAR(20) NULL,double_field FLOAT NULL,bool_field BOOLEAN NULL"
            + ",uint_field INTEGER NULL,bloqueo BOOLEAN NOT NULL,empty_relation VARCHAR(15) NULL,int_field INTEGER NULL);"
        )
        driver = flsqlite.FLSQLITE()

        self.assertEqual(sql, driver.sqlCreateTable(cursor.metadata()))
        self.assertEqual(sql2, driver.sqlCreateTable(cursor.metadata(), False))

    def test_basic_3(self) -> None:
        """Basics test 3."""

        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        conn_ = cursor.db()
        ret = conn_.driver().recordInfo2("fltest")
        self.assertEqual(["id", "uint", True, 0, None, None, True], ret["id"])
        self.assertTrue(conn_.alterTable(cursor.metadata()))

    def test_mismatched(self) -> None:
        """Test mismatched table."""

        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor2 = pnsqlcursor.PNSqlCursor("fltest3")

        metadata = cursor.metadata()
        metadata2 = cursor2.metadata()

        self.assertFalse(cursor.db().driver().mismatchedTable("fltest", metadata))
        self.assertTrue(cursor.db().driver().mismatchedTable("fltest2", metadata))
        self.assertTrue(cursor.db().driver().mismatchedTable("fltest3", metadata))
        metadata.removeFieldMD("date_field")
        self.assertTrue(cursor.db().driver().mismatchedTable("fltest", metadata))
        self.assertTrue(cursor.db().driver().mismatchedTable("fltest", metadata2))
        self.assertFalse(cursor.db().driver().mismatchedTable("fltest3", metadata2))

    def test_invalid_metadata(self) -> None:
        """Test invalid metadata."""
        from pineboolib.application.database import pnsqlcursor

        with self.assertRaises(Exception):
            pnsqlcursor.PNSqlCursor("fltest6").metadata()

        pnsqlcursor.PNSqlCursor("fltest").metadata()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Fldateedit module."""

# -*- coding: utf-8 -*-
import datetime
from typing import Union, Optional

from PyQt6 import QtCore, QtWidgets  # type: ignore
from pineboolib.q3widgets import qdateedit
from pineboolib.application.utils.date_conversion import convert_to_qdate
from pineboolib.application.types import Date


class FLDateEdit(qdateedit.QDateEdit):
    """FLDateEdit class."""

    valueChanged: QtCore.pyqtSignal = QtCore.pyqtSignal()
    DMY: str = "dd-MM-yyyy"
    _parent: QtWidgets.QWidget

    def __init__(self, parent: "QtWidgets.QWidget", name: str) -> None:
        """Inicialize."""

        super().__init__(parent, name)
        self.DMY = "dd-MM-yyyy"  # pylint: disable=invalid-name
        self.setMinimumWidth(90)
        # self.setMaximumWidth(90)
        self._parent = parent

    def setOrder(self, order: str) -> None:
        """Set order."""
        self.setDisplayFormat(order)

    def getDate(self) -> str:
        """Return date."""
        return super().getDate()

    def setDate(  # type: ignore [override] # noqa F821
        self, date: Union[str, datetime.date, Date, QtCore.QDate] = ""
    ) -> None:
        """Set Date."""
        new_date: Union[str, datetime.date, Date, QtCore.QDate, None] = (
            None if isinstance(date, QtCore.QDate) and date.isNull() else date
        )

        super().setDate(
            QtCore.QDate.fromString(str("01-01-2000"), self.DMY)
            if new_date in (None, "NAN", "")
            else convert_to_qdate(date)
        )
        self.setStyleSheet("color: black")

    date: Optional[str] = property(getDate, setDate)  # type: ignore [assignment] # noqa F821



"""Test_flmssql module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.plugins.sql import flpyodbc


class TestFLMSSql(unittest.TestCase):
    """TestFLSqlite Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Basics test 1."""

        driver = flpyodbc.FLPYODBC()

        self.assertEqual(driver.formatValueLike("bool", "true", False), "=0")
        self.assertEqual(driver.formatValueLike("date", "27-01-2020", True), "LIKE '%%2020-01-27'")

        self.assertEqual(driver.formatValue("bool", "false", True), "0")
        self.assertEqual(driver.formatValue("time", "", True), "")

        self.assertEqual(driver.setType("String", 20), "VARCHAR(20)")
        self.assertEqual(driver.setType("sTring", 0), "VARCHAR")
        self.assertEqual(driver.setType("Double"), "DECIMAL")
        self.assertEqual(driver.setType("Bool"), "BIT")
        self.assertEqual(driver.setType("DATE"), "DATE")
        self.assertEqual(driver.setType("pixmap"), "TEXT")
        self.assertEqual(driver.setType("bytearray"), "NVARCHAR")
        self.assertEqual(driver.setType("timestamp"), "DATETIME2")

    def test_basic_2(self) -> None:
        """Basics test 1."""
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        sql = (
            "CREATE TABLE fltest (id PRIMARY KEY,string_field VARCHAR NULL,date_field DATE NULL,time_field TIME NULL,double_field"
            + " DECIMAL(8,2) NULL,bool_field BIT NULL,uint_field BIGINT NULL,"
            + "bloqueo BIT NOT NULL,empty_relation VARCHAR(15) NULL,int_field INT NULL)"
        )

        driver = flpyodbc.FLPYODBC()
        print(driver.sqlCreateTable(cursor.metadata(), False))
        self.assertEqual(sql, driver.sqlCreateTable(cursor.metadata(), False))

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""
ISqlCursor module.
"""

from PyQt6 import QtCore  # type: ignore[import]


from pineboolib.interfaces import cursoraccessmode


from typing import Any, Optional, Dict, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.acls import pnboolflagstate  # noqa: F401 # pragma: no cover
    from pineboolib.application.database import (  # noqa : F401
        pnbuffer,  # noqa: F401
        pncursortablemodel,  # noqa: F401
    )  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata import (  # noqa : F401 # pragma: no cover
        pntablemetadata,
        pnrelationmetadata,
        pnaction,
    )
    from pineboolib.interfaces import iconnection  # pragma: no cover
    from sqlalchemy.ext import declarative  # type: ignore [import]


class ICursorPrivate(QtCore.QObject):
    """ICursorPrivate class."""

    """
    Buffer with a cursor record.

    According to the FLSqlCursor :: Mode access mode set for the cusor, this buffer will contain
    the active record of said cursor ready to insert, edit, delete or navigate.
    """

    buffer_: Optional["pnbuffer.PNBuffer"] = None

    """
    Copia del buffer.

    Aqui se guarda una copia del FLSqlCursor::buffer_ actual mediante el metodo FLSqlCursor::updateBufferCopy().
    """
    _buffer_copy: Optional["pnbuffer.PNBuffer"] = None

    """
    Metadatos de la tabla asociada al cursor.
    """
    metadata_: Optional["pntablemetadata.PNTableMetaData"]

    """
    Mantiene el modo de acceso actual del cursor, ver FLSqlCursor::Mode.
    """
    mode_access_ = -1

    """
    Cursor relacionado con este.
    """
    cursor_relation_: Optional["ISqlCursor"]

    """
    Relación que determina como se relaciona con el cursor relacionado.
    """
    relation_: Optional["pnrelationmetadata.PNRelationMetaData"]

    """
    Esta bandera cuando es TRUE indica que se abra el formulario de edición de regitros en
    modo edición, y cuando es FALSE se consulta la bandera FLSqlCursor::browse. Por defecto esta
    bandera está a TRUE
    """
    edition_: bool

    """
    Esta bandera cuando es TRUE y la bandera FLSqlCuror::edition es FALSE, indica que se
    abra el formulario de edición de registro en modo visualización, y cuando es FALSE no hace
    nada. Por defecto esta bandera está a TRUE
    """
    browse_: bool
    browse_states_: "pnboolflagstate.PNBoolFlagStateList"

    """
    Filtro principal para el cursor.

    Este filtro persiste y se aplica al cursor durante toda su existencia,
    los filtros posteriores, siempre se ejecutaran unidos con 'AND' a este.
    """
    # self.d._model.where_filters["main-filter"] = None

    """
    Accion asociada al cursor, esta accion pasa a ser propiedad de FLSqlCursor, que será el
    encargado de destruirla
    """
    action_: "pnaction.PNAction"

    """
    Cuando esta propiedad es TRUE siempre se pregunta al usuario si quiere cancelar
    cambios al editar un registro del cursor.
    """
    _ask_for_cancel_changes: bool

    """
    Indica si estan o no activos los chequeos de integridad referencial
    """
    _activated_check_integrity: bool

    """
    Indica si estan o no activas las acciones a realiar antes y después del Commit
    """
    _activated_commit_actions: bool

    """
    Contexto de ejecución de scripts.

    El contexto de ejecución será un objeto formulario el cual tiene asociado un script.
    Ese objeto formulario corresponde a aquel cuyo origen de datos es este cursor.
    El contexto de ejecución es automáticamente establecido por las clases FLFormXXXX.
    """
    ctxt_: Any

    """
    Cronómetro interno
    """
    timer_: Optional[QtCore.QTimer]

    """
    Cuando el cursor proviene de una consulta indica si ya se han agregado al mismo
    la definición de los campos que lo componen
    """
    populated_: bool

    """
    Cuando el cursor proviene de una consulta contiene la sentencia sql
    """
    _is_query: bool

    """
    Cuando el cursor proviene de una consulta contiene la clausula order by
    """
    _query_order_by: str

    """
    Base de datos sobre la que trabaja
    """
    db_: Optional["iconnection.IConnection"]

    """
    Pila de los niveles de transacción que han sido iniciados por este cursor
    """
    _transactions_opened: List[int]

    """
    Filtro persistente para incluir en el cursor los registros recientemente insertados aunque estos no
    cumplan los filtros principales. Esto es necesario para que dichos registros sean válidos dentro del
    cursor y así poder posicionarse sobre ellos durante los posibles refrescos que puedan producirse en
    el proceso de inserción. Este filtro se agrega a los filtros principales mediante el operador OR.
    """
    _persistent_filter: Optional[str]

    """
    Cursor propietario
    """
    cursor_: Optional["ISqlCursor"]

    """
    Nombre del cursor
    """
    cursor_name_: str

    """
    Orden actual
    """
    sort_: str
    """
    Auxiliares para la comprobacion de riesgos de bloqueos
    """
    _in_loop_risk_locks: bool
    _in_risks_locks: bool

    """
    Para el control de acceso dinámico en función del contenido de los registros
    """

    acl_table_: Dict[str, Any] = {}
    _ac_perm_table = None
    _acos_permanent_backup_table: Dict[str, str] = {}
    _acos_table: List[str] = []
    _acos_backup_table: Dict[str, str] = {}
    _acos_cond_name: Optional[str] = None
    _acos_cond: int
    _acos_cond_value = None
    _last_at = None
    _acl_done = False
    _id_ac = 0
    _id_acos = 0
    _id_cond = 0
    id_ = "000"
    _init_orm: bool = False

    """ Uso interno """
    _is_system_table: bool
    # rawValues_: bool

    _md5_tuples: str

    _count_ref_cursor: int

    _model: "pncursortablemodel.PNCursorTableModel"

    edition_states_: "pnboolflagstate.PNBoolFlagStateList"
    _current_changed = QtCore.pyqtSignal(int)
    _id_acl: str

    _currentregister: int

    def __init__(
        self, cursor_: "ISqlCursor", action_name: str, db_: "iconnection.IConnection"
    ) -> None:
        """
        Initialize the private part of the cursor.
        """

        super().__init__()

    def __del__(self) -> None:
        """
        Delete instance values.
        """

        pass  # pragma: no cover

    def msgBoxWarning(self, msg: str, throw_exception: bool = False) -> None:
        """Return msgbox if an error exists."""

        pass  # pragma: no cover

    def needUpdate(self) -> bool:  # type: ignore [empty-body]
        """Indicate if the cursor needs to be updated."""

        pass  # pragma: no cover

    def undoAcl(self) -> None:
        """Delete restrictions according to access control list."""

        pass  # pragma: no cover

    def doAcl(self) -> None:
        """Create restrictions according to access control list."""

        pass  # pragma: no cover


class ISqlCursor(QtCore.QObject):
    """
    Abstract class for PNSqlCursor.
    """

    """
    signals:
    """

    """
    Indica que se ha cargado un nuevo buffer
    """
    newBuffer = QtCore.pyqtSignal()

    """
    Indica ha cambiado un campo del buffer, junto con la señal se envía el nombre del campo que
    ha cambiado.
    """
    bufferChanged = QtCore.pyqtSignal(str)

    """
    Indica que se ha actualizado el cursor
    """
    cursorUpdated = QtCore.pyqtSignal()

    """
    Indica que se ha elegido un registro, mediante doble clic sobre él o bien pulsando la tecla Enter
    """
    recordChoosed = QtCore.pyqtSignal()

    """
    Indica que la posicion del registro activo dentro del cursor ha cambiado
    """
    currentChanged = QtCore.pyqtSignal(int)

    """
    Indica que se ha realizado un commit automático para evitar bloqueos
    """
    autoCommit = QtCore.pyqtSignal()

    """
    Indica que se ha realizado un commitBuffer
    """
    bufferCommited = QtCore.pyqtSignal()

    """
    Indica que se ha cambiado la conexión de base de datos del cursor. Ver changeConnection
    """
    connectionChanged = QtCore.pyqtSignal()

    """
    Indica que se ha realizado un commit
    """
    commited = QtCore.pyqtSignal()

    Insert = cursoraccessmode.CursorAccessMode.Insert
    Edit = cursoraccessmode.CursorAccessMode.Edit
    Del = cursoraccessmode.CursorAccessMode.Del
    Browse = cursoraccessmode.CursorAccessMode.Browse
    Value = 0
    RegExp = 1
    Function = 2

    private_cursor: "ICursorPrivate"

    _selection: Optional[QtCore.QItemSelectionModel] = None

    _iter_current: Optional[int]

    _action: Optional["pnaction.PNAction"] = None

    _name: str

    transactionBegin: QtCore.pyqtSignal = QtCore.pyqtSignal()
    transactionEnd: QtCore.pyqtSignal = QtCore.pyqtSignal()
    transactionRollback: QtCore.pyqtSignal = QtCore.pyqtSignal()

    _cursor_model: "declarative.DeclarativeMeta"

    _is_delegate_commit: bool
    _last_delegate_commit_result: bool
    _persistent_filter_deletegate: Optional[str]

    def __init__(
        self,
        name: Optional[str] = None,
        conn_or_autopopulate: Union[bool, str] = True,
        connection_name_or_db: Union[str, "iconnection.IConnection"] = "default",
        cursor_relation: Optional["ISqlCursor"] = None,
        relation: Optional["pnrelationmetadata.PNRelationMetaData"] = None,
        parent=None,
    ) -> None:
        """Create cursor."""
        super().__init__()

    def init(self, name: str, autopopulate, cusor_relation, relation) -> None:
        """Initialize cursor."""
        pass  # pragma: no cover

    def conn(self) -> "iconnection.IConnection":  # type: ignore [empty-body]
        """Retrieve connection object."""
        pass  # pragma: no cover

    def table(self) -> str:
        """Retrieve table name."""
        return ""  # pragma: no cover

    def setName(self, name, autop) -> None:
        """Set cursor name."""
        pass  # pragma: no cover

    def metadata(self) -> "pntablemetadata.PNTableMetaData":  # type: ignore [empty-body]
        """Get table metadata for this cursor table."""
        pass  # pragma: no cover

    def currentRegister(self) -> int:  # type: ignore [empty-body]
        """Get current row number."""
        pass  # pragma: no cover

    def modeAccess(self) -> int:  # type: ignore [empty-body]
        """Get current access mode."""
        pass  # pragma: no cover

    def filter(self) -> str:
        """Get SQL filter as a string."""
        return ""

    def mainFilter(self) -> str:  # type: ignore [empty-body]
        """Get SQL Main filter as a string."""
        pass  # pragma: no cover

    def action(self) -> Optional["pnaction.PNAction"]:
        """Get action object."""
        pass  # pragma: no cover

    def actionName(self) -> str:  # type: ignore [empty-body]
        """Get action name."""
        pass  # pragma: no cover

    def setAction(self, action) -> bool:  # type: ignore [empty-body]
        """Set Action object."""
        pass  # pragma: no cover

    def setMainFilter(self, filter: str, do_refresh: bool = True) -> None:
        """Set Main filter for this cursor."""
        pass  # pragma: no cover

    def setModeAccess(self, mode_access) -> None:
        """Set Access mode for the cursor."""
        pass  # pragma: no cover

    def connectionName(self) -> str:  # type: ignore [empty-body]
        """Get current connection name."""
        pass  # pragma: no cover

    def setValueBuffer(self, field_name: str, value: Any) -> None:
        """Set Value on the cursor buffer."""
        pass  # pragma: no cover

    def valueBuffer(self, field_name: str, with_no_value: bool = False) -> Any:
        """Get value from cursor buffer."""
        return False  # pragma: no cover

    def fetchLargeValue(self, value) -> Optional[str]:
        """Fetch from fllarge."""
        pass  # pragma: no cover

    def valueBufferCopy(self, field_name: str, with_no_value: bool = False) -> Any:
        """Get original value on buffer."""
        pass  # pragma: no cover

    def setEdition(self, value, flag=None) -> None:
        """Set edit mode."""
        pass  # pragma: no cover

    def restoreEditionFlag(self, flag) -> None:
        """Restore edit flag."""
        pass  # pragma: no cover

    def setBrowse(self, value, flag=None) -> None:
        """Set browse mode."""
        pass  # pragma: no cover

    def restoreBrowseFlag(self, flag) -> None:
        """Restore browse flag."""
        pass  # pragma: no cover

    # def meta_model(self) -> Any:
    #    """Get sqlAlchemy model."""
    #    pass  # pragma: no cover

    def setContext(self, context=None) -> None:
        """Set script execution context."""
        pass  # pragma: no cover

    def context(self) -> Any:
        """Get script execution context."""
        pass  # pragma: no cover

    def fieldDisabled(self, field_name) -> bool:  # type: ignore [empty-body]
        """Get if field is disabled."""
        pass  # pragma: no cover

    def inTransaction(self) -> bool:  # type: ignore [empty-body]
        """Return if transaction is in progress."""
        pass  # pragma: no cover

    def transaction(self, lock=False) -> bool:  # type: ignore [empty-body]
        """Open transaction."""
        pass  # pragma: no cover

    def rollback(self) -> bool:  # type: ignore [empty-body]
        """Rollback transaction."""
        pass  # pragma: no cover

    def commit(self, notify=True) -> bool:  # type: ignore [empty-body]
        """Commit transaction."""
        pass  # pragma: no cover

    def size(self) -> int:  # type: ignore [empty-body]
        """Get current cursor size in rows."""
        pass  # pragma: no cover

    def openFormInMode(self, mode: int, wait: bool = True, cont: bool = True) -> None:
        """Open record form in specified mode."""
        pass  # pragma: no cover

    def isNull(self, field_name) -> bool:  # type: ignore [empty-body]
        """Get if field is null."""
        pass  # pragma: no cover

    def updateBufferCopy(self) -> None:
        """Refresh buffer copy."""
        pass  # pragma: no cover

    def isModifiedBuffer(self) -> bool:  # type: ignore [empty-body]
        """Get if buffer is modified."""
        pass  # pragma: no cover

    def setAskForCancelChanges(self, value) -> None:
        """Activate dialog for asking before closing."""
        pass  # pragma: no cover

    def setActivatedCheckIntegrity(self, value) -> None:
        """Activate integrity checks."""
        pass  # pragma: no cover

    def activatedCheckIntegrity(self) -> bool:  # type: ignore [empty-body]
        """Get integrity check state."""
        pass  # pragma: no cover

    def setActivatedCommitActions(self, value) -> None:
        """Activate before/after commit."""
        pass  # pragma: no cover

    def activatedCommitActions(self) -> bool:  # type: ignore [empty-body]
        """Get before/after commit status."""
        pass  # pragma: no cover

    def cursorRelation(self) -> Optional["ISqlCursor"]:
        """Get cursor relation."""
        pass  # pragma: no cover

    def relation(self) -> Optional["pnrelationmetadata.PNRelationMetaData"]:
        """Get relation."""
        pass  # pragma: no cover

    def setUnLock(self, field_name, value) -> None:
        """Set unlock field."""
        pass  # pragma: no cover

    def isLocked(self) -> bool:  # type: ignore [empty-body]
        """Get if record is locked."""
        pass  # pragma: no cover

    def buffer(self) -> "pnbuffer.PNBuffer":  # type: ignore [empty-body]
        """Get buffer object."""
        pass  # pragma: no cover

    def bufferCopy(self) -> "pnbuffer.PNBuffer":  # type: ignore [empty-body]
        """Get buffer copy."""
        pass  # pragma: no cover

    def setNull(self, name) -> None:
        """Set field to null."""
        pass  # pragma: no cover

    def db(self) -> "iconnection.IConnection":  # type: ignore [empty-body]
        """Return database object."""
        pass  # pragma: no cover

    def curName(self) -> str:  # type: ignore [empty-body]
        """Get cursor name."""
        pass  # pragma: no cover

    def filterAssoc(self, field_name, table_metadata=None) -> Optional[str]:
        """Retrieve filter for associated field."""
        pass  # pragma: no cover

    def calculateField(self, field_name) -> bool:  # type: ignore [empty-body]
        """Return the result of a field calculation."""
        pass  # pragma: no cover

    def model(self) -> "pncursortablemodel.PNCursorTableModel":  # type: ignore [empty-body]
        """Get sqlAlchemy model."""
        pass  # pragma: no cover

    def selection(self) -> Optional["QtCore.QItemSelectionModel"]:
        """Get selection."""
        pass  # pragma: no cover

    def at(self) -> int:  # type: ignore [empty-body]
        """Get row number."""
        pass  # pragma: no cover

    def isValid(self) -> bool:  # type: ignore [empty-body]
        """Return if cursor is valid."""
        pass  # pragma: no cover

    def refresh(self, field_name=None) -> None:
        """Refresh cursor."""
        pass  # pragma: no cover

    def refreshBuffer(self) -> bool:  # type: ignore [empty-body]
        """Refresh buffer."""
        pass  # pragma: no cover

    def setEditMode(self) -> bool:  # type: ignore [empty-body]
        """Set cursor in edit mode."""
        pass  # pragma: no cover

    def seek(self, i, relative=None, emite=None) -> bool:  # type: ignore [empty-body]
        """Move cursor without fetching."""
        pass  # pragma: no cover

    def next(self, emite=True) -> bool:  # type: ignore [empty-body]
        """Get next row."""
        pass  # pragma: no cover

    def moveby(self, pos) -> bool:  # type: ignore [empty-body]
        """Move cursor down "pos" rows."""
        pass  # pragma: no cover

    def prev(self, emite=True) -> bool:  # type: ignore [empty-body]
        """Get previous row."""
        pass  # pragma: no cover

    def move(self, row) -> bool:  # type: ignore [empty-body]
        """Move cursor to row number."""
        pass  # pragma: no cover

    def first(self, emite=True) -> bool:  # type: ignore [empty-body]
        """Move cursor to first row."""
        pass  # pragma: no cover

    def last(self, emite=True) -> bool:  # type: ignore [empty-body]
        """Move cursor to last row."""
        pass  # pragma: no cover

    def select(self, _filter=None, sort=None) -> bool:  # type: ignore [empty-body]
        """Perform SQL Select."""
        pass  # pragma: no cover

    def setSort(self, filter) -> None:
        """Set sorting order."""
        pass  # pragma: no cover

    def insertRecord(self, wait: bool = True) -> None:
        """Open form in insert mode."""
        pass  # pragma: no cover

    def editRecord(self, wait: bool = True) -> None:
        """Open form in edit mode."""
        pass  # pragma: no cover

    def browseRecord(self, wait: bool = True) -> None:
        """Open form in browse mode."""
        pass  # pragma: no cover

    def deleteRecord(self, wait: bool = True) -> None:
        """Delete record."""
        pass  # pragma: no cover

    def copyRecord(self) -> None:
        """Copy record."""
        pass  # pragma: no cover

    def chooseRecord(self) -> None:
        """Emit chooseRecord."""
        pass  # pragma: no cover

    def setForwardOnly(self, value) -> None:
        """Set forward only."""
        pass  # pragma: no cover

    def commitBuffer(self, emite=True, check_locks=False) -> bool:  # type: ignore [empty-body]
        """Commit current buffer to db."""
        pass  # pragma: no cover

    def commitBufferCursorRelation(self) -> bool:  # type: ignore [empty-body]
        """Commit buffer from cursor relation."""
        pass  # pragma: no cover

    def transactionLevel(self) -> int:  # type: ignore [empty-body]
        """Get number of nested transactions."""
        pass  # pragma: no cover

    def transactionsOpened(self) -> List[str]:  # type: ignore [empty-body]
        """Return if any transaction is open."""
        pass  # pragma: no cover

    def rollbackOpened(self, count=-1, msg=None) -> None:
        """Return if in rollback."""
        pass  # pragma: no cover

    def commitOpened(self, count=-1, msg=None) -> None:
        """Return if in commit."""
        pass  # pragma: no cover

    def checkIntegrity(self, show_error: bool = True) -> bool:  # type: ignore [empty-body]
        """Return check integrity result."""
        pass  # pragma: no cover

    def checkRisksLocks(self, terminate: bool = False) -> bool:  # type: ignore [empty-body]
        """Return risks locks result."""

        pass  # pragma: no cover

    def msgCheckIntegrity(self) -> str:  # type: ignore [empty-body]
        """Return msg check integrity."""

        pass  # pragma: no cover

    def aqWasDeleted(self) -> bool:  # type: ignore [empty-body]
        """Indicate if the cursor has been deleted."""

        pass  # pragma: no cover

    def concurrencyFields(self) -> List[str]:  # type: ignore [empty-body]
        """Return list of concurrency fields."""

        pass  # pragma: no cover

    def setFilter(self, _filter: str = "") -> None:
        """
        Specify the cursor filter.

        @param _filter. Text string with the filter to apply.
        """

        pass  # pragma: no cover

    # def field(self, name: str) -> Optional["pnbuffer.FieldStruct"]:
    #    """
    #    Return a specified FieldStruct of the buffer.
    #    """

    #    pass # pragma: no cover

    def curFilter(self) -> str:
        """
        Return the actual filter.

        @return actual filter.
        """

        return ""

    def sort(self) -> str:
        """
        Choose the order of the main columns.

        @return sort order.
        """

        return ""

    def id(self) -> str:
        """
        Return cursor identifier.
        """

        return ""

    def primaryKey(self) -> str:
        """
        Return the primary cursor key.

        @return primary key field name.
        """

        return ""

    def clear_buffer(self) -> None:
        """Clear buffer."""

        pass  # pragma: no cover

    def bufferIsNull(self, field_name: str) -> bool:
        """Return if buffer is null."""

        return False

    def doCommitBuffer(self, emite: bool = True) -> bool:
        """Lanza llamada sengun proceda el deletateCommit o commitBuffer del cursorRelation."""

        return False

    def doCommit(self) -> bool:
        """Lanza commit del cursor o reposiciona el cusor, sengun proceda."""

        return False

    def useDelegateCommit(self) -> bool:
        """Retorna si se cumplen las condiciones para usar delegateCommit."""

        return False

    def setPersistentFilterDelegate(self, filter: str) -> None:
        """Añade a persistent filter datos de delegate."""

        pass

    def restorePersistentFilterBeforeDelegate(self):
        """Restaura persistent filter despues de hacer commit."""

        pass

    def setInitOrm(self, initorm: bool) -> None:
        """Set init orm."""

        pass



"""Flspinbox module."""
# -*- coding: utf-8 -*-

from pineboolib.q3widgets import qspinbox

from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # type: ignore[import] # noqa: F401 # pragma: no cover


class FLSpinBox(qspinbox.QSpinBox):
    """FLSpinBox class."""

    def __init__(self, parent: Optional["QtWidgets.QWidget"] = None) -> None:
        """Inicialize."""
        super().__init__(parent)
        # editor()setAlignment(Qt::AlignRight);

    def setMaxValue(self, value: int) -> None:
        """Set maximum value."""

        self.setMaximum(value)

    def getValue(self) -> int:
        """Return actual value."""

        return super().value()

    def setValue(self, val: int) -> None:
        """Set a value."""
        super().setValue(val)

    value: Any = property(getValue, setValue)  # type: ignore
    text: Any = value



"""Object_class module."""

from PyQt6 import QtCore  # type: ignore[import]

from typing import Set, Tuple, Any
from pineboolib.application import connections


class ObjectClass(object):
    """ObjectClass class."""

    _class_connections: Set[Tuple]
    signal_test = QtCore.pyqtSignal(str, QtCore.QObject)

    def module_connect(self, sender: Any, signal: str, receiver: Any, slot: str) -> None:
        """Connect two objects."""

        signal_slot = connections.connect(sender, signal, receiver, slot, caller=self)
        if not signal_slot:
            return

        if not hasattr(self, "_class_connections"):
            self._class_connections = set([])

        self._class_connections.add(signal_slot)

    def module_disconnect(self, sender: Any, signal: str, receiver: Any, slot: str) -> None:
        """Disconnect two objects."""

        # print(" > > > disconnect:", self)

        signal_slot = connections.disconnect(sender, signal, receiver, slot)
        if not signal_slot:
            return

        if hasattr(self, "_class_connections"):
            for conn_ in self._class_connections:
                # PyQt6-Stubs misses signal.signal
                if (
                    conn_[0].signal == getattr(signal_slot[0], "signal")
                    and conn_[1].__name__ == signal_slot[1].__name__
                ):
                    self._class_connections.remove(conn_)
                    break



"""options module."""

import optparse
from typing import Optional, List


def parse_options(custom_argv: Optional[List] = None) -> "optparse.Values":
    """Load and parse options."""

    parser = optparse.OptionParser()
    parser.add_option(
        "-l",
        "--load",
        dest="project",
        help="load projects/PROJECT.xml and run it",
        metavar="PROJECT",
    )
    parser.add_option(
        "-s",
        "--silentconn",
        dest="connection",
        help="connect to database with user and password.",
        metavar="user:passwd:driver_alias@host:port/database",
    )

    parser.add_option(
        "-c",
        "--call",
        dest="call_function",
        help="call a specific function",
        metavar="module.pub_function",
    )

    parser.add_option(
        "-f",
        "--folder",
        dest="flfiles_folder",
        help="use flfiles folder instead flfiles table",
        metavar="FLFILES_PATH",
    )

    parser.add_option(
        "-u",
        "--update_flfiles",
        action="store_true",
        dest="update_flfiles",
        default=False,
        help="Update flfiles table from flfiles_folder",
    )

    parser.add_option(
        "-x",
        "--exit",
        action="store_true",
        dest="quit_after_call",
        help="exit after call a specific function",
        default=False,
    )

    parser.add_option(
        "-e",
        "--external_modules",
        dest="external_modules",
        help="use external folder which contains python modules to load.",
        metavar="EXTERNAL_MODULES_PATH",
    )

    parser.add_option(
        "-v", "--verbose", action="count", default=0, help="increase verbosity level"
    )  # default a 2 para ver los logger.info, 1 no los muestra
    parser.add_option("-q", "--quiet", action="count", default=0, help="decrease verbosity level")
    parser.add_option(
        "--profile-time",
        action="store_true",
        dest="enable_profiler",
        default=False,
        help="Write profile information about CPU load after running",
    )
    parser.add_option(
        "--trace-debug",
        action="store_true",
        dest="trace_debug",
        default=False,
        help="Write lots of trace information to stdout",
    )
    parser.add_option(
        "--trace-loggers",
        dest="trace_loggers",
        default="",
        help="Comma separated list of modules to enable TRACE output",
    )
    parser.add_option(
        "--log-time",
        action="store_true",
        dest="log_time",
        default=False,
        help="Add timestamp to logs",
    )
    parser.add_option(
        "--log-sql",
        action="store_true",
        dest="log_sql",
        default=False,
        help="Write sql log to stdout",
    )
    parser.add_option(
        "--trace-signals",
        action="store_true",
        dest="trace_signals",
        default=False,
        help="Wrap up every signal, connect and emit, and give useful tracebacks",
    )
    parser.add_option("-a", "--action", dest="action", help="load action", metavar="ACTION")
    parser.add_option(
        "--no-python-cache",
        action="store_true",
        dest="no_python_cache",
        default=False,
        help="Always translate QS to Python",
    )
    parser.add_option(
        "--preload",
        action="store_true",
        dest="preload",
        default=False,
        help="Load everything. Then exit. (Populates Pineboo cache)",
    )
    parser.add_option(
        "--force-load",
        dest="forceload",
        default=None,
        metavar="ACTION",
        help="Preload actions containing string ACTION without caching. Useful to debug pythonyzer",
    )
    parser.add_option(
        "--dgi", dest="dgi", default="qt", help="Change the gdi mode by default", metavar="DGI"
    )
    parser.add_option(
        "--dgi_parameter",
        dest="dgi_parameter",
        help="Change the gdi mode by default",
        metavar="DGIPARAMETER",
    )
    parser.add_option(
        "--dbadmin",
        action="store_true",
        dest="enable_dbadmin",
        default=False,
        help="Enables DBAdmin mode",
    )
    parser.add_option(
        "--quick",
        action="store_true",
        dest="enable_quick",
        default=False,
        help="Enables Quick mode",
    )

    parser.add_option(
        "--pool-sql-preping",
        action="store_true",
        dest="enable_preping",
        default=False,
        help="Enables SQL connections preping",
    )

    parser.add_option(
        "--no-x",
        action="store_false",
        dest="enable_gui",
        default=True,
        help="Disables graphical interface",
    )
    parser.add_option(
        "--main_form", dest="main_form", help="Change the main form", metavar="MAINFORMPARAMETER"
    )

    parser.add_option(
        "--version",
        dest="pineboo_version",
        help="Returns pineboo version",
        action="store_true",
        default=False,
    )

    parser.add_option(
        "--no-acl",
        action="store_false",
        dest="enable_acls",
        default=True,
        help="Disable acls loads",
    )

    parser.add_option(
        "--no-interactive-gui",
        action="store_false",
        dest="enable_interactive_gui",
        default=True,
        help="interactiveGUI return empty value",
    )

    parser.add_option(
        "--no-qsa-exceptions",
        action="store_false",
        dest="enable_call_exceptions",
        default=True,
        help="disable qsa call exceptions",
    )

    parser.add_option(
        "--external",
        dest="external",
        help="load external python module",
        metavar="MODULE_DIR",
    )

    parser.add_option(
        "--project-name",
        dest="project_name",
        help="Project name",
        metavar="PROJECT_NAME",
    )

    parser.add_option(
        "--parse-project",
        action="store_false",
        dest="parse_project_on_init",
        default=False,
        help="parse all project on init",
    )

    parser.add_option(
        "-i",
        "--omit-no_python",
        action="store_true",
        dest="omit_no_python_tags",
        default=False,
        help="Omit tags no_python",
    )

    if custom_argv is None:
        (options, args) = parser.parse_args()
    else:
        (options, args) = parser.parse_args(custom_argv)

    # ---- OPTIONS POST PROCESSING -----
    if options.forceload:
        options.no_python_cache = True
        options.preload = True

    options.loglevel = 30 + (options.quiet - options.verbose) * 5
    options.debug_level = 200  # 50 - (options.quiet - options.verbose) * 25

    return options



"""Flcodbar module."""

# # -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtGui, QtSvg  # type: ignore

from pineboolib.core.utils.utils_base import load2xml

from pineboolib import logging

import barcode  # type: ignore # pip3 install python-barcode
from typing import Dict, Any, cast, Optional

LOGGER = logging.get_logger(__name__)

BARCODE_ANY: int = 0
BARCODE_EAN: int = 1
BARCODE_EAN_8: int = 2
BARCODE_EAN_13: int = 3
BARCODE_EAN_14: int = 4
BARCODE_UPC: int = 5
BARCODE_UPC_A: int = 6
BARCODE_JAN: int = 7
BARCODE_ISBN: int = 8
BARCODE_ISBN_10: int = 9
BARCODE_ISBN_13: int = 10
BARCODE_ISSN: int = 11
BARCODE_39: int = 12
BARCODE_128: int = 13
BARCODE_PZN: int = 14
BARCODE_ITF: int = 15
BARCODE_GS1: int = 16
BARCODE_GTIN: int = 17


class FLCodBar(object):
    """FLCodBar class."""

    barcode: Dict[str, Any]
    _pixmap: Optional[QtGui.QPixmap]
    _pixmap_error: QtGui.QPixmap

    def __init__(
        self,
        value: Any = None,
        type_: int = BARCODE_128,
        margin: int = 10,
        scale: float = 1.0,
        cut: float = 1.0,
        rotation: int = 0,
        text_flag: bool = False,
        fg: QtGui.QColor = cast(QtGui.QColor, QtCore.Qt.GlobalColor.black),
        bg: QtGui.QColor = cast(QtGui.QColor, QtCore.Qt.GlobalColor.white),
        res: int = 72,
    ) -> None:
        """Inicialize."""

        dict_ = {"barcode": "python-barcode"}
        from pineboolib.application.utils.check_dependencies import check_dependencies

        check_dependencies(dict_)
        self._pixmap_error = QtGui.QPixmap()
        self.barcode = {}
        self.barcode["value"] = ""
        self._pixmap = None

        if value in [None, 0]:
            self.reading_std_out = False
            self.writing_std_out = False
            self.fillDefault(self.barcode)
        else:
            if isinstance(value, str):
                self.reading_std_out = False
                self.writing_std_out = False
                self.barcode["value"] = value
                self.barcode["type"] = type_
                self.barcode["margin"] = margin
                self.barcode["scale"] = scale
                self.barcode["cut"] = cut
                self.barcode["rotation"] = rotation
                self.barcode["text"] = text_flag
                self.barcode["fg"] = fg
                self.barcode["bg"] = bg
                self.barcode["valid"] = False
                self.barcode["res"] = res
            elif isinstance(value, int):
                raise ValueError("Not supported")
            elif isinstance(value, FLCodBar):
                self._copyBarCode(value.barcode, self.barcode)

    def pixmap(self) -> QtGui.QPixmap:
        """Return pixmap barcode."""

        self._createBarcode()

        if not self._pixmap:
            self.barcode["valid"] = False
            return self.pixmapError()

        return self._pixmap

    def pixmapError(self) -> QtGui.QPixmap:
        """Return empy pixmap barcode."""

        return self._pixmap_error

    def value(self) -> Any:
        """Return value."""
        return self.barcode["value"]

    def type_(self) -> int:
        """Return type."""

        return self.barcode["type"]

    def margin(self) -> int:
        """Return margin."""

        return self.barcode["margin"]

    def scale(self) -> float:
        """Return scale."""

        return self.barcode["scale"]

    def cut(self) -> float:
        """Return cut."""
        return self.barcode["cut"]

    def text(self) -> str:
        """Return text."""

        return self.barcode["text"]

    def rotation(self) -> int:
        """Return rotation."""

        return self.barcode["rotation"]

    def fg(self) -> QtGui.QColor:
        """Return fore ground color."""

        return self.barcode["fg"]

    def bg(self) -> QtGui.QColor:
        """Return back ground color."""
        return self.barcode["bg"]

    def setData(self, data: Dict[str, Any]) -> None:
        """Set barcode data."""

        self.barcode = data

    def validBarcode(self) -> bool:
        """Return if is a valid barcode."""

        return self.barcode["valid"]

    def setCaption(self, caption: str) -> None:
        """Set caption barcode."""

        self.barcode["caption"] = caption

    def caption(self) -> str:
        """Return caption barcode."""

        return self.barcode["caption"]

    def setValue(self, value: Any) -> None:
        """Set value."""

        self.barcode["value"] = value

    def setType(self, type_: int) -> None:
        """Set type."""
        self.barcode["type"] = type_

    def setMargin(self, margin: int) -> None:
        """Set margin size."""
        self.barcode["margin"] = margin

    def setScale(self, scale: float) -> None:
        """Set scale."""
        self.barcode["scale"] = scale

    def setCut(self, cut: float) -> None:
        """Set cut size."""
        self.barcode["cut"] = cut

    def setText(self, text: str) -> None:
        """Set text size."""
        self.barcode["text"] = text

    def setRotation(self, rotation: int) -> None:
        """Set rotation size."""
        self.barcode["rotation"] = rotation

    def setFg(self, fore_ground: QtGui.QColor) -> None:
        """Set fore ground color."""
        self.barcode["fg"] = fore_ground

    def setBg(self, back_ground: QtGui.QColor) -> None:
        """Set back ground color."""
        self.barcode["bg"] = back_ground

    def setRes(self, res: int) -> None:
        """Set resolution."""

        self.barcode["res"] = res

    def data(self) -> Dict[str, Any]:
        """Return barcode data."""
        return self.barcode

    def fillDefault(self, data: Dict[str, Any] = {}) -> None:
        """Fill with default values."""

        if not data:
            data = {}

        data["bg"] = "white"
        data["fg"] = "black"
        data["margin"] = 10
        data["text"] = True
        data["value"] = "1234567890"
        data["type"] = BARCODE_39
        data["scale"] = 1.0
        data["cut"] = 1.0
        data["rotation"] = 0
        data["caption"] = "Static"
        data["valid"] = False
        data["res"] = 72
        self.setData(data)

    def cleanUp(self) -> None:
        """Clean barcode data."""

        if self._pixmap:
            self._pixmap.scaled(0, 0)
        self._pixmap_error.scaled(0, 0)

    def nameToType(self, name: str) -> int:
        """Return barcode  name from type."""
        name = name.lower()
        if name == "any":
            return BARCODE_ANY
        elif name == "ean":
            return BARCODE_EAN
        elif name == "ean-8":
            return BARCODE_EAN_8
        elif name == "ean-13":
            return BARCODE_EAN_13
        elif name == "ean-14":
            return BARCODE_EAN_14
        elif name == "upc":
            return BARCODE_UPC
        elif name == "upc-a":
            return BARCODE_UPC_A
        elif name == "jan":
            return BARCODE_JAN
        elif name == "isbn":
            return BARCODE_ISBN
        elif name == "isbn-10":
            return BARCODE_ISBN_10
        elif name == "isbn-13":
            return BARCODE_ISBN_13
        elif name == "issn":
            return BARCODE_ISSN
        elif name == "code39":
            return BARCODE_39
        elif name == "code128":
            return BARCODE_128
        elif name == "pzn":
            return BARCODE_PZN
        elif name == "itf":
            return BARCODE_ITF
        elif name == "gs1":
            return BARCODE_GS1
        elif name == "gtin":
            return BARCODE_GTIN
        else:
            LOGGER.warning(
                "Formato no soportado (%s)\nSoportados: %s." % (name, barcode.PROVIDED_BARCODES)
            )
            return BARCODE_ANY

    def typeToName(self, type_: int) -> str:
        """Return barcode type from barcode name."""
        if type_ == BARCODE_ANY:
            return "ANY"
        elif type_ == BARCODE_EAN:
            return "EAN"
        elif type_ == BARCODE_EAN_8:
            return "EAN-8"
        elif type_ == BARCODE_EAN_13:
            return "EAN-13"
        elif type_ == BARCODE_EAN_14:
            return "EAN-14"
        elif type_ == BARCODE_UPC:
            return "UPC"
        elif type_ == BARCODE_UPC_A:
            return "UPC-A"
        elif type_ == BARCODE_JAN:
            return "JAN"
        elif type_ == BARCODE_ISBN:
            return "ISBN"
        elif type_ == BARCODE_ISBN_10:
            return "ISBN-10"
        elif type_ == BARCODE_ISBN_13:
            return "ISBN-13"
        elif type_ == BARCODE_ISSN:
            return "ISSN"
        elif type_ == BARCODE_39:
            return "Code39"
        elif type_ == BARCODE_128:
            return "Code128"
        elif type_ == BARCODE_PZN:
            return "PZN"
        elif type_ == BARCODE_ITF:
            return "ITF"
        elif type_ == BARCODE_GS1:
            return "GS1"
        elif type_ == BARCODE_GTIN:
            return "GTIN"
        else:
            return "ANY"

    def _createBarcode(self) -> None:
        """Create barcode."""
        if self.barcode["value"] == "":
            return
        if self.barcode["type"] == BARCODE_ANY:
            LOGGER.warning("Usando %s por defecto" % self.typeToName(BARCODE_128))
            self.barcode["type"] = BARCODE_128

        type_ = self.typeToName(self.barcode["type"])
        value_ = self.barcode["value"]
        bg_ = self.barcode["bg"]
        fg_ = self.barcode["fg"]
        if not isinstance(self.barcode["bg"], str):
            bg_ = QtGui.QColor(self.barcode["bg"]).name()

        if not isinstance(self.barcode["fg"], str):
            fg_ = QtGui.QColor(self.barcode["fg"]).name()

        margin_ = self.barcode["margin"] / 10

        render_options: Dict[str, Any] = {}
        render_options["module_width"] = 0.6
        render_options["module_height"] = 10
        render_options["background"] = bg_.lower()
        render_options["foreground"] = fg_.lower()
        render_options["font_size"] = 8
        render_options["write_text"] = self.barcode["text"]
        render_options["text_distance"] = 35
        render_options["quiet_zone"] = margin_
        if self.barcode["text"]:
            render_options["text"] = value_
        else:
            render_options["text"] = " "

        bar_code = barcode.get_barcode_class(type_.lower())
        try:
            bar_ = bar_code("%s" % value_)
        except Exception:
            bar_ = bar_code("000000000000")

        svg = bar_.render(render_options)
        xml_svg = load2xml(svg.decode("utf-8")).getroot()
        xwidth, xheight = xml_svg.get("width"), xml_svg.get("height")
        if xwidth and xheight:
            svg_w = 3.779 * float(xwidth[0:6])
            svg_h = 3.779 * float(xheight[0:6])
        else:
            LOGGER.warning("width or height missing")
            svg_w = 0.0
            svg_h = 0.0
        self._pixmap = QtGui.QPixmap(int(svg_w), int(svg_h))
        render = QtSvg.QSvgRenderer(svg)
        self._pixmap.fill(QtCore.Qt.GlobalColor.transparent)
        painter = QtGui.QPainter(self._pixmap)
        render.render(painter, QtCore.QRectF(0, 0, svg_w * 3.4, svg_h * 3.4))

        if self._pixmap.isNull():
            self.barcode["valid"] = False
        else:
            if self.barcode["scale"] != 1.0:
                width_scaled = self.barcode["x"] * self.barcode["scale"]
                height_scaled = self.barcode["y"] * self.barcode["scale"]
                self._pixmap = self._pixmap.scaled(width_scaled, height_scaled)

            self.barcode["x"] = self._pixmap.width()
            self.barcode["y"] = self._pixmap.height()

            self.barcode["valid"] = True

    def _copyBarCode(self, source: Dict[str, Any], dest: Dict[str, Any]) -> None:
        """Copy a new barcode from another."""

        dest["value"] = source["value"]
        dest["type"] = source["type"]
        dest["margin"] = source["margin"]
        dest["scale"] = source["scale"]
        dest["cut"] = source["cut"]
        dest["rotation"] = source["rotation"]
        dest["text"] = source["text"]
        dest["caption"] = source["caption"]
        dest["valid"] = source["valid"]
        dest["fg"] = source["fg"]
        dest["bg"] = source["bg"]
        if "x" in source.keys():
            dest["x"] = source["x"]
            dest["y"] = source["y"]
        dest["res"] = source["res"]



# -*- coding: utf-8 -*-
"""
AQSobjectsFactory Module.

This module provides the different classes and AQS functions to be used in the module scripts.
"""
from pineboolib.core.utils.utils_base import is_deployed as __is_deployed
from pineboolib.core.utils.utils_base import is_library as __is_library

# AQSObjects
from pineboolib.fllegacy.aqsobjects.aqsettings import AQSettings  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsqlquery import AQSqlQuery  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsqlcursor import AQSqlCursor  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqutil import AQUtil  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsql import AQSql  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsmtpclient import AQSmtpClient  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqs import AQS  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqboolflagstate import (  # noqa: F401
    AQBoolFlagState,
    AQBoolFlagStateList,
)
from pineboolib.fllegacy.aqsobjects.aqformdb import aq_form_db as AQFormDB  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqssproject import AQSSProject  # noqa: F401
from pineboolib.fllegacy.aqsobjects.aqsbuttongroup import AQSButtonGroup  # noqa: F401


if (
    not __is_deployed() and not __is_library()
):  # FIXME: No module named 'xml.sax.expatreader' in deploy
    from pineboolib.fllegacy.aqsobjects.aqods import (  # noqa: F401
        AQOdsGenerator,
        AQOdsSpreadSheet,
        AQOdsSheet,
        AQOdsRow,
    )
    from pineboolib.fllegacy.aqsobjects.aqods import AQOdsStyle, AQOdsImage  # noqa: F401
    from pineboolib.fllegacy.aqsobjects.aqods import aq_ods_color as AQOdsColor  # noqa: F401



"""Checkbox module."""

# -*- coding: utf-8 -*-

from PyQt6 import QtWidgets  # type: ignore[import]
from pineboolib.q3widgets.qwidget import QWidget


class CheckBox(QWidget):
    """CheckBox class."""

    _label: QtWidgets.QLabel
    _cb: QtWidgets.QCheckBox

    def __init__(self) -> None:
        """Inicialize."""

        super().__init__()

        self._label = QtWidgets.QLabel(self)
        self._cb = QtWidgets.QCheckBox(self)
        spacer = QtWidgets.QSpacerItem(
            1, 1, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum
        )
        _lay = QtWidgets.QHBoxLayout()
        _lay.addWidget(self._cb)
        _lay.addWidget(self._label)
        _lay.addSpacerItem(spacer)
        self.setLayout(_lay)

    # def __setattr__(self, name: str, value: Any) -> None:
    #    """Set an attribute."""

    #    if name == "text":
    #        self._label.setText(str(value))
    #    elif name == "checked":
    #        self._cb.setChecked(value)

    # def __getattr__(self, name: str) -> Any:
    #    """Return an attribute."""

    #    if name == "checked":
    #        return self._cb.isChecked()

    def getText(self) -> str:
        """Return text label."""

        return self._label.text()

    def setText(self, value: str) -> None:
        """Set text label."""

        self._label.setText(value)

    def getChecked(self) -> bool:
        """Return if checked."""

        return self._cb.isChecked()

    def setChecked(self, value: bool) -> None:
        """Set checked."""

        self._cb.setChecked(value)

    text: str = property(getText, setText)  # type: ignore [assignment] # noqa: F821
    checked: bool = property(getChecked, setChecked)  # type: ignore [assignment] # noqa: F821



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Fltest5_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Fltest5(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Fltest5 class."""

    __tablename__ = "fltest5"

    # --- Metadata --->
    legacy_metadata = {
        "name": "fltest5",
        "alias": "Test table",
        "fields": [
            {
                "name": "id",
                "alias": "ID",
                "pk": True,
                "type": "serial",
                "null": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "idmodulo",
                "alias": "Id. del Módulo",
                "type": "string",
                "length": 15,
                "null": False,
            },
            {
                "name": "idarea",
                "alias": "Id. del Área",
                "type": "string",
                "length": 15,
                "relations": [{"card": "M1", "table": "fltest4", "field": "idarea", "delC": True}],
                "null": False,
            },
            {"name": "string_timestamp", "alias": "String timestamp", "type": "timestamp"},
            {"name": "uint_field", "alias": "Unsigned int field", "type": "uint"},
            {"name": "my_json", "alias": "json field", "type": "json"},
            {
                "name": "version",
                "alias": "Versión",
                "type": "string",
                "length": 3,
                "regexp": r"^(\d{1,2}(,\d{1,2})*)?$",
                "null": True,
                "default": "0.0",
                "editable": False,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    id = sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True)
    idmodulo = sqlalchemy.Column("idmodulo", sqlalchemy.String(15))
    idarea = sqlalchemy.Column("idarea", sqlalchemy.String(15))
    string_timestamp = sqlalchemy.Column("string_timestamp", sqlalchemy.DateTime)
    uint_field = sqlalchemy.Column("uint_field", sqlalchemy.BigInteger)
    my_json = sqlalchemy.Column("my_json", sqlalchemy.types.JSON)


# <--- Fields ---



"""
IManager Module.
"""
from typing import Any, Dict, Optional, Union, List, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.application.database import pnsqlquery  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata import pntablemetadata  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata import pnfieldmetadata  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata import pnrelationmetadata  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata import pnaction  # pragma: no cover
    from pineboolib.interfaces import iconnection  # pragma: no cover
    from xml.etree import ElementTree  # noqa: F401 # pragma: no cover
    from PyQt6 import QtXml  # type: ignore[import] # noqa: F401 # pragma: no cover
#     import pineboolib.application.database.pnconnection
#     import pineboolib.application.metadata.pnfieldmetadata
#     import pineboolib.application.metadata.pntablemetadata
#     import pineboolib.application.metadata.pnrelationmetadata
#     import pineboolib.application.metadata.pnaction


class IManager(object):
    """
    Abstract class for FLManager.
    """

    list_tables_: List[str]  # Lista de las tablas de la base de datos, para optimizar lecturas
    dict_key_metadata_: Dict[
        str, str
    ]  # Diccionario de claves de metadatos, para optimizar lecturas
    cache_metadata_: Dict[
        str, "pntablemetadata.PNTableMetaData"
    ]  # Caché de metadatos, para optimizar lecturas
    _cache_action: Dict[
        str, "pnaction.PNAction"
    ]  # Caché de definiciones de acciones, para optimizar lecturas
    # Caché de metadatos de talblas del sistema para optimizar lecturas

    db_: "iconnection.IConnection"  # Base de datos a utilizar por el manejador

    __doc__: str
    buffer_: None

    def __init__(self, *args) -> None:
        """Create manager."""
        return None  # pragma: no cover

    def action(self, name: str) -> "pnaction.PNAction":  # "pnaction.PNAction"
        """Retrieve action object by name."""
        raise Exception("must be implemented")  # pragma: no cover

    def alterTable(self, metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Issue an alter table to db."""
        return False  # pragma: no cover

    def checkMetaData(self, mtd1, mtd2) -> bool:
        """Validate MTD against DB."""
        return False  # pragma: no cover

    def cleanupMetaData(self) -> None:
        """Clean up MTD."""
        return None  # pragma: no cover

    def createSystemTable(self, name: str) -> bool:
        """Create named system table."""
        return False  # pragma: no cover

    def createTable(self, name_or_metadata) -> Optional["pntablemetadata.PNTableMetaData"]:
        """Create new table."""
        return None  # pragma: no cover

    def existsTable(self, name: str, cache: bool = False) -> bool:
        """Check if table does exist in db."""
        return False  # pragma: no cover

    def fetchLargeValue(self, ref_key: str) -> Optional[str]:
        """Fetch from fllarge."""
        return None  # pragma: no cover

    def finish(self) -> None:
        """Finish?."""
        return None  # pragma: no cover

    def formatAssignValue(self, *args) -> str:
        """Format value for DB update."""
        return ""  # pragma: no cover

    def formatAssignValueLike(self, *args) -> str:
        """Format value for DB "LIKE" statement."""
        return ""  # pragma: no cover

    def formatValue(self, fmd_or_type: str, value: Any, upper: bool = False) -> str:
        """Format value for DB."""
        return ""  # pragma: no cover

    def formatValueLike(
        self,
        fmd_or_type: Union["pnfieldmetadata.PNFieldMetaData", str],
        value: Any,
        upper: bool = False,
    ) -> str:
        """Format value for DB LIKE."""
        return ""  # pragma: no cover

    def init(self) -> None:
        """Initialize this object."""
        return None  # pragma: no cover

    def initCount(self) -> int:
        """Track number of inits."""
        return 0  # pragma: no cover

    def isSystemTable(self, name: str) -> bool:
        """Return if given name is a system table."""
        return False  # pragma: no cover

    def loadTables(self) -> None:
        """Load tables."""
        return None  # pragma: no cover

    def metadata(
        self, name_or_xml, quick: bool = False
    ) -> Optional["pntablemetadata.PNTableMetaData"]:  # PNTableMetaData"
        """Retrieve table metadata by table name."""
        return None  # pragma: no cover

    def metadataField(
        self, field: "ElementTree.Element", vvisible: bool = False, ededitable: bool = False
    ) -> Optional["pnfieldmetadata.PNFieldMetaData"]:  # "PNFieldMetaData"
        """Retrieve field metadata."""
        raise Exception("must be implemented")  # pragma: no cover

    def metadataRelation(
        self, relation: Union["QtXml.QDomElement", "ElementTree.Element"]
    ) -> Optional["pnrelationmetadata.PNRelationMetaData"]:  # "PNRelationMetaData"
        """Retrieve relationship."""
        raise Exception("must be implemented")  # pragma: no cover

    def query(
        self, name: str, parent: Optional["pnsqlquery.PNSqlQuery"]
    ) -> Optional["pnsqlquery.PNSqlQuery"]:  # "PNSqlQuery"
        """Create query."""
        return None  # pragma: no cover

    def storeLargeValue(self, mtd, large_value: str) -> Optional[str]:
        """Store value in fllarge."""
        return None  # pragma: no cover



"""
test_pnrelationmetadata Module.
"""

import unittest
from pineboolib.loader.main import init_testing
from pineboolib import application


class TestPNRelationMetaData(unittest.TestCase):
    """TestPNRelationMetaData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """PNRelationMetaData."""

        mtd = application.PROJECT.conn_manager.manager().metadata("flgroups")
        if mtd is None:
            raise Exception
        rel_1 = mtd.relation("idgroup", "idgroup", "flusers")
        rel_2 = mtd.relation("idgroup", "idgroup", "flacs")
        if rel_1 is None:
            raise Exception
        if rel_2 is None:
            raise Exception

        self.assertEqual(rel_1.field(), "idgroup")
        self.assertEqual(rel_2.foreignField(), "idgroup")
        self.assertEqual(rel_2.foreignTable(), "flacs")
        self.assertEqual(rel_2.deleteCascade(), False)
        self.assertEqual(rel_1.cardinality(), "1M")

    def test_assigment(self) -> None:
        """Test multiples relaitos asigment to a field."""

        from pineboolib.application.metadata import pnrelationmetadata

        mtd = application.PROJECT.conn_manager.manager().metadata("flmodules")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("version")
            self.assertTrue(field)
            if field is not None:
                relation0 = pnrelationmetadata.PNRelationMetaData("tabla0", "campo0", "M1")
                relation1 = pnrelationmetadata.PNRelationMetaData("tabla1", "campo1", "1M")
                relation2 = pnrelationmetadata.PNRelationMetaData("tabla2", "campo2", "1M")
                relation3 = pnrelationmetadata.PNRelationMetaData("tabla3", "campo3", "M1")
                relation4 = pnrelationmetadata.PNRelationMetaData("tabla4", "campo4", "M1")
                self.assertFalse(field.relationM1())
                self.assertFalse(field.relationList())
                field.addRelationMD(relation0)
                self.assertFalse(relation0 in field.relationList())
                self.assertEqual(relation0, field.relationM1())
                field.addRelationMD(relation1)
                self.assertTrue(relation1 in field.relationList())
                self.assertNotEqual(relation1, field.relationM1())
                field.addRelationMD(relation2)
                self.assertTrue(relation2 in field.relationList())
                field.addRelationMD(relation3)
                self.assertNotEqual(relation3, field.relationM1())
                self.assertFalse(relation3 in field.relationList())
                field.addRelationMD(relation4)
                self.assertNotEqual(relation4, field.relationM1())
                self.assertFalse(relation4 in field.relationList())
                self.assertEqual([relation1, relation2], field.relationList())


class TestCreatePNRelationMetaData(unittest.TestCase):
    """TestCreatePNRelationMetaData Class."""

    def test_basic(self) -> None:
        """PNRelationMetaData."""

        from pineboolib.application.metadata import pnrelationmetadata

        mtd = application.PROJECT.conn_manager.manager().metadata("flgroups")
        if mtd is None:
            raise Exception
        rel_1 = mtd.relation("idgroup", "idgroup", "flusers")
        if rel_1 is None:
            raise Exception("Relation is empty!.")
        rel_2 = pnrelationmetadata.PNRelationMetaData(rel_1)
        rel_3 = pnrelationmetadata.PNRelationMetaData(
            "flgroups", "idgroup", "M1", True, True, False
        )

        self.assertEqual(rel_2.foreignField(), "idgroup")
        self.assertEqual(rel_2.foreignTable(), "flusers")

        rel_3.setField("new_field")

        self.assertEqual(rel_3.field(), "new_field")
        self.assertEqual(rel_3.foreignField(), "idgroup")
        self.assertEqual(rel_3.cardinality(), "M1")
        self.assertEqual(rel_3.deleteCascade(), True)
        self.assertEqual(rel_3.updateCascade(), True)
        self.assertEqual(rel_3.checkIn(), False)


if __name__ == "__main__":
    unittest.main()



"""Test_dateedit module."""

import unittest


class TestDateEdit(unittest.TestCase):
    """TestDateEdit Class."""

    def test_all(self) -> None:
        """Test DateEdit."""

        from pineboolib.qsa import qsa

        date_edit = qsa.DateEdit()
        date_edit.label = "Prueba label"
        date_edit.date = qsa.Date("2019-01-10")
        self.assertEqual(str(date_edit.date), "2019-01-10T00:00:00")
        self.assertEqual(date_edit.label, "Prueba label")



"""
IFieldMetaData module.
"""
from typing import List, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces.itablemetadata import ITableMetaData  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata.pnrelationmetadata import (  # noqa: F401
        PNRelationMetaData,  # noqa: F401
    )  # noqa: F401 # pragma: no cover


class IFieldMetaData:
    """
    Abastract class for FieldMetaData.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Create new FieldMetaData."""
        return  # pragma: no cover

    def __len__(self) -> int:
        """Get size of field."""
        return 0  # pragma: no cover

    def addRelationMD(self, relation) -> None:
        """Add relation M1 or 1M."""
        return  # pragma: no cover

    def alias(self) -> str:
        """Get alias for this field."""
        return ""  # pragma: no cover

    def allowNull(self) -> bool:
        """Determine if this field allos NULLs."""
        return False  # pragma: no cover

    def associatedField(self) -> Optional["IFieldMetaData"]:
        """Return associated field."""
        return None  # pragma: no cover

    def associatedFieldFilterTo(self) -> str:
        """Return associated field filter in sql string format."""
        return ""  # pragma: no cover

    def associatedFieldName(self) -> Optional[str]:
        """Return associated field name."""
        return None  # pragma: no cover

    def defaultValue(self) -> Optional[Union[str, bool]]:
        """Return field default value."""
        return None  # pragma: no cover

    def editable(self) -> bool:
        """Get if field is editable."""
        return False  # pragma: no cover

    def formatAssignValue(self, field_name: str, value: int, upper: bool) -> str:
        """Format a value for this field."""
        return ""  # pragma: no cover

    def generated(self) -> bool:
        """Get if field is computed."""
        return False  # pragma: no cover

    def getIndexOptionsList(self, name: str) -> Optional[int]:
        """Get list of options."""
        return None  # pragma: no cover

    def hasOptionsList(self) -> bool:
        """Return if this field has list of options."""
        return False  # pragma: no cover

    def inicializeFLFieldMetaData(self, other) -> None:
        """Initialize."""
        return  # pragma: no cover

    def inicializeNewFLFieldMetaData(
        self,
        name: str,
        alias: str,
        allow_null: bool,
        is_primary_key: bool,
        ttype: str,
        length_: int = 0,
        calculated: bool = False,
        visible: bool = True,
        editable: bool = True,
        part_integer: int = 4,
        part_decimal: int = 0,
        is_index: bool = False,
        is_unique: bool = False,
        coun: bool = False,
        default_value: Optional[str] = None,
        out_transaction: bool = False,
        regular_exp: Optional[str] = None,
        visible_grib: bool = True,
        generated: bool = True,
        is_compound_key: bool = False,
    ) -> None:
        """Initialize."""
        return  # pragma: no cover

    def isCompoundKey(self) -> bool:
        """Return if this field is part of CK."""
        return False  # pragma: no cover

    def isPrimaryKey(self) -> bool:
        """Return if this field is PK."""
        return False  # pragma: no cover

    def length(self) -> int:
        """Return field size."""
        return 0  # pragma: no cover

    def metadata(self) -> Optional["ITableMetaData"]:
        """Return table metadata for this field."""
        return None  # pragma: no cover

    def name(self) -> str:
        """Get name of this field."""
        return ""  # pragma: no cover

    def optionsList(self) -> List[str]:
        """Get list of options for this field."""
        return []  # pragma: no cover

    def outTransaction(self) -> bool:
        """Return if this field should be updated outside of transaction."""
        return False  # pragma: no cover

    def partDecimal(self) -> int:
        """Return the amount of digits after dot when formatting numbers."""
        return 0  # pragma: no cover

    def partInteger(self) -> int:
        """Return amount of digits before decimal dot."""
        return 0  # pragma: no cover

    def regExpValidator(self) -> Optional[str]:
        """Validate regexp."""
        return None  # pragma: no cover

    def relationM1(self) -> Optional["PNRelationMetaData"]:
        """Return M1 relationship in this field."""
        return None  # pragma: no cover

    def setAssociatedField(self, relation_or_name, field: str) -> None:
        """Set new associated field."""
        return  # pragma: no cover

    def setEditable(self, editable: bool) -> None:
        """Set if this field should be editable."""
        return  # pragma: no cover

    def setFullyCalculated(self, calculated: bool) -> None:
        """Set if this field should be fully calculated."""
        return  # pragma: no cover

    def setMetadata(self, metadata) -> None:
        """Set TableMetadata for this field."""
        return  # pragma: no cover

    def setOptionsList(self, options_list: str) -> None:
        """Set option list for this field."""
        return  # pragma: no cover

    def setTrimed(self, trimed: bool) -> None:
        """Set if this field should be trimed."""
        return  # pragma: no cover

    def setVisible(self, visible: bool) -> None:
        """Set if this field should be visible."""
        return  # pragma: no cover

    def type(self) -> str:
        """Return field type."""
        return ""  # pragma: no cover

    def visible(self) -> bool:
        """Get if this field should be visible in UI."""
        return False  # pragma: no cover

    def visibleGrid(self) -> bool:
        """Get if this field should be visible in grids."""
        return False  # pragma: no cover



"""Test bytearray module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa
from pineboolib.application.database.orm.tests import fixture_path
from datetime import datetime


class TestByteArray(unittest.TestCase):
    """TestByteArray Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test basic."""

        image_path = fixture_path("tux.png")
        session = qsa.thread_session_new()
        session.begin()
        obj = qsa.orm.flupdates()
        file_ = open(image_path, "rb")
        obj.auxbin = file_.read()
        file_.close()
        obj.fecha = datetime.now().date()
        obj.hora = datetime.now().time()
        obj.nombre = "prueba"
        obj.modulesdef = ""
        obj.filesdef = ""

        obj.shaglobal = ""
        obj.save()
        session.commit()
        qsa.thread_session_free()

    def test_basic2(self) -> None:
        """Test basic 2."""

        util = qsa.FLUtil()
        session = qsa.session()
        session.begin()
        image_path = fixture_path("tux.png")

        file_ = open(image_path, "rb")
        data = file_.read()
        file_.close()

        self.assertTrue(
            util.sqlInsert(
                "flupdates",
                ["fecha", "hora", "nombre", "modulesdef", "filesdef", "shaglobal", "auxbin"],
                [
                    datetime.now().date(),
                    datetime.now().time(),
                    "prueba2",
                    "",
                    "",
                    "",
                    data,
                ],
            )
        )
        session.rollback()
        session.close()

    def test_basic3(self) -> None:
        """Test basic 3."""

        util = qsa.FLUtil()
        session = qsa.session()
        session.begin()
        sql = "INSERT INTO flupdates(id, fecha, hora , nombre, modulesdef, filesdef, shaglobal, actual, auxbin, auxtxt)"
        sql += " VALUES (666,'%s','%s','prueba3','x','y','z', False,'aa:123', 'tt')" % (
            datetime.now().date(),
            datetime.now().time(),
        )
        result = util.execSql(sql)
        self.assertTrue(result is not False)
        session.rollback()
        session.close()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""Progressdialogmanager module."""
from typing import Any, List
from PyQt6 import QtCore, QtWidgets  # type: ignore


class ProgressDialogManager(object):
    """ProgressDailogManager class."""

    progress_dialog_stack: List[Any]

    def __init__(self):
        """Inicialize."""
        self.progress_dialog_stack = []

    def create(self, title: str, steps: int, id_: str = "default") -> Any:
        """Create new ProgressDialog."""

        pd_widget = QtWidgets.QProgressDialog(
            str(title), str(QtWidgets.QApplication.translate("scripts", "Cancelar")), 0, steps
        )
        if pd_widget is not None:
            self.progress_dialog_stack.append(pd_widget)
            pd_widget.setObjectName(id_)
            pd_widget.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
            pd_widget.setWindowTitle(str(title))

            pd_widget.show()
            # pd_widget.setMinimumDuration(100)
            QtWidgets.QApplication.processEvents()

        return pd_widget

    def destroy(self, id_: str = "default") -> None:
        """Destroy a specific progress dialog."""

        for dialog in self.progress_dialog_stack:
            if dialog.objectName() == id_:
                self.progress_dialog_stack.remove(dialog)
                del dialog
                return

    def setProgress(self, step_number: int, id_: str = "default") -> None:
        """Set progress into a specific prores dialog."""

        for dialog in self.progress_dialog_stack:
            if dialog.objectName() == id_:
                dialog.setValue(step_number)
                return

    def setLabelText(self, label: str, id_: str) -> None:
        """Set label text to a specific progres dialog."""

        for dialog in self.progress_dialog_stack:
            if dialog.objectName() == id_:
                dialog.setLabelText(str(label))
                return

    def setTotalSteps(self, total_steps: int, id_: str) -> None:
        """Set total steps to a specific proress dialog."""

        for dialog in self.progress_dialog_stack:
            if dialog.objectName() == id_:
                dialog.setRange(0, total_steps)
                return



"""Test_flformsearchdb module."""

from pineboolib.fllegacy.tests import fixture_path

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa
from pineboolib.core.utils import utils_base

from PyQt6 import QtWidgets  # type: ignore[import]


class TestFLFormsearchDB(unittest.TestCase):
    """TestFLFormsearchDB Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_initialization(self) -> None:
        """Test flformrecord cursor assignment"""

        import os

        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa.sys.loadModules(path, False))

        form_search_1 = qsa.FLFormSearchDB("bancos")
        self.assertEqual(form_search_1.cursor().metadata().name(), "bancos")
        self.assertEqual(form_search_1.action().name(), "bancos")
        self.assertEqual(form_search_1.parent(), QtWidgets.QApplication.activeModalWidget())

        parent_2 = QtWidgets.QWidget()

        form_search_2 = qsa.FLFormSearchDB("clientes", parent_2)
        self.assertEqual(form_search_2.cursor().metadata().name(), "clientes")
        self.assertEqual(form_search_2.action().name(), "clientes")
        self.assertEqual(form_search_2.parent(), parent_2)

        parent_3 = QtWidgets.QWidget()
        cur_3 = qsa.FLSqlCursor("proveedores")
        print(1, cur_3)
        form_search_3 = qsa.FLFormSearchDB(cur_3, "proveedores", parent_3)
        print(2, cur_3)
        self.assertEqual(form_search_3.cursor().metadata().name(), "proveedores")
        self.assertEqual(form_search_3.action().name(), "proveedores")
        self.assertEqual(form_search_3.parent(), parent_3)
        self.assertEqual(form_search_3.cursor(), cur_3)

        parent_4 = None
        cur_4 = qsa.FLSqlCursor("proveedores")
        form_search_4 = qsa.FLFormSearchDB(cur_4, "proveedores", parent_4)
        self.assertEqual(form_search_4.cursor().metadata().name(), "proveedores")
        self.assertEqual(form_search_4.action().name(), "proveedores")
        self.assertEqual(form_search_4.parent(), QtWidgets.QApplication.activeModalWidget())
        self.assertEqual(form_search_4.cursor(), cur_4)

        self.assertEqual(qsa.sys.interactiveGUI(), "Pineboo")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()


class TestNoFLFormsearchDB(unittest.TestCase):
    """TestNoFLFormsearchDB Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        utils_base.FORCE_DESKTOP = False
        init_testing()

    def test_no_gui(self) -> None:
        """Test FlformSearch call with no GUI"""

        self.assertEqual(qsa.sys.interactiveGUI(), "Pinebooapi")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        utils_base.FORCE_DESKTOP = True
        finish_testing()



"""Connection Module."""

import optparse
from pineboolib.loader import projectconfig

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.database import pnconnection

DEFAULT_SQLITE_CONN = projectconfig.ProjectConfig(
    database="pineboo.sqlite3", type="SQLite3 (SQLITE3)"
)
IN_MEMORY_SQLITE_CONN = projectconfig.ProjectConfig(
    database=":memory:", type="SQLite3 (SQLITE3)", username="memory_user"
)


def config_dbconn(options: "optparse.Values") -> Optional["projectconfig.ProjectConfig"]:
    """Obtain a config connection from a file."""

    if options.project:  # FIXME: --project debería ser capaz de sobreescribir algunas opciones
        if not options.project.endswith(".xml"):
            options.project += ".xml"
        prj_name = options.project
        try:
            return projectconfig.ProjectConfig(load_xml=prj_name)
        except projectconfig.PasswordMismatchError:
            # If fails without password, ignore the exception so the stack is cleaned.
            # This avoids seeing two exceptions if password is wrong.
            pass

        import getpass

        password = getpass.getpass()
        return projectconfig.ProjectConfig(load_xml=prj_name, project_password=password)

    if options.connection:
        return projectconfig.ProjectConfig(connstring=options.connection)

    return None


def connect_to_db(config: "projectconfig.ProjectConfig") -> "pnconnection.PNConnection":
    """Try connect a database with projectConfig data."""
    if config.database is None:
        raise ValueError("database not set")
    if config.type is None:
        raise ValueError("type not set")

    from pineboolib.application.database import pnconnection

    port = int(config.port) if config.port else None
    connection = pnconnection.PNConnection(
        config.database,
        config.host or "",
        port or 0,
        config.username or "",
        config.password or "",
        config.type,
    )
    return connection



"""Qeventloop module."""

from PyQt6 import QtCore  # type: ignore[import]


class QEventLoop(QtCore.QEventLoop):
    """QEventLoop class."""

    def exitLoop(self) -> None:
        """Call exit loop."""
        super().exit()

    def enterLoop(self) -> None:
        """Call exec_ loop."""
        super().exec()



"""
Manage Qt Signal-Slot connections.
"""


from PyQt6 import QtCore, QtWidgets  # type: ignore[import]
from pineboolib import logging

import inspect
import weakref
import re


from typing import Callable, Any, Dict, Tuple, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    import types  # pragma: no cover
    from pineboolib.qsa import formdbwidget  # noqa F401 # pragma: no cover
    from pineboolib.qsa import object_class  # noqa F401 # pragma: no cover


LOGGER = logging.get_logger(__name__)


class ProxySlot:
    """
    Proxies a method so it doesn't need to be resolved on connect.
    """

    PROXY_FUNCTIONS: Dict[str, "Callable"] = {}

    def __init__(
        self, remote_fn: "types.MethodType", receiver: "QtCore.QObject", slot: str
    ) -> None:
        """Create a proxy for a method."""
        self.key = "%r.%r->%r" % (remote_fn, receiver, slot)
        if self.key not in self.PROXY_FUNCTIONS:
            self.PROXY_FUNCTIONS[self.key] = proxy_fn(
                weakref.WeakMethod(remote_fn), weakref.ref(receiver)
            )
        self.proxy_function = self.PROXY_FUNCTIONS[self.key]

    def getProxyFn(self) -> Callable:
        """Retrieve internal proxy function."""
        return self.proxy_function


def get_expected_args_num(inspected_function: "Callable") -> int:
    """Inspect function to get how many arguments expects."""
    expected_args = inspect.getfullargspec(inspected_function)[0]
    args_num = len(expected_args)

    if args_num and expected_args[0] == "self":
        args_num -= 1

    return args_num


def get_expected_kwargs(inspected_function: "Callable") -> bool:
    """Inspect a function to get if expects keyword args."""
    return True if inspect.getfullargspec(inspected_function)[2] else False


def proxy_fn(weak_ref_method: "weakref.WeakMethod", weak_ref: "weakref.ref") -> "Callable":
    """Create a proxied function, so it does not hold the garbage collector."""

    def function(*args: Any, **kwargs: Any) -> Optional[Any]:
        function_method = weak_ref_method()
        if not function_method:
            return None  # pragma: no cover
        ref = weak_ref()
        if not ref:
            return None  # pragma: no cover

        args_num = get_expected_args_num(function_method)

        if args_num:
            args_list = list(args)
            while len(args_list) < args_num:
                args_list.append(None)
            args = tuple(args_list)

            return function_method(*args[0:args_num], **kwargs)
        else:
            return function_method()

    return function


def slot_done(function: Callable, signal: "QtCore.pyqtSignal") -> Callable:
    """Create a fake slot for QS connects."""

    def new_fn(*args: Any, **kwargs: Any) -> Any:
        # PyQt6-Stubs seems to miss QtCore.pyqtSignal.name (also, this seems to be internal)
        original_signal_name: str = getattr(signal, "signal")

        # Este parche es para evitar que las conexiones de un clicked de error de cantidad de argumentos.
        # En Eneboo se esperaba que signal no contenga argumentos
        if original_signal_name == "2clicked(bool)":
            args = tuple()  # pragma: no cover
        # args_num = get_expected_args_num(fn)
        try:
            return function(*args, **kwargs) if get_expected_kwargs(function) else function(*args)
        except Exception:  # pragma: no cover
            LOGGER.exception("Error trying to create a connection")  # pragma: no cover

        return False  # pragma: no cover

    return new_fn


def connect(
    sender: "QtWidgets.QWidget",
    signal: str,
    receiver: "QtCore.QObject",
    slot: str,
    caller: Optional[Union["formdbwidget.FormDBWidget", "object_class.ObjectClass"]] = None,
) -> Optional[Tuple["QtCore.pyqtSignal", Callable]]:
    """Connect signal to slot for QSA."""

    # Parameters example:
    # caller: <clientes.FormInternalObj object at 0x7f78b5c230f0>
    # sender: <pineboolib.q3widgets.qpushbutton.QPushButton object at 0x7f78b4de1af0>
    # signal: 'clicked()'
    # receiver: <clientes.FormInternalObj object at 0x7f78b5c230f0>
    # slot: 'iface.buscarContacto()'

    if caller is not None:
        LOGGER.trace("* * * Connect:: %s %s %s %s %s", caller, sender, signal, receiver, slot)
    else:
        LOGGER.trace(
            "? ? ? Connect:: %s %s %s %s", sender, signal, receiver, slot
        )  # pragma: no cover
    signal_slot = solve_connection(sender, signal, receiver, slot)

    if not signal_slot:
        return None
    # http://pyqt.sourceforge.net/Docs/PyQt4/qt.html#ConnectionType-enum
    # conntype =
    #    QtCore.Qt.ConnectionType.QueuedConnection,
    #    QtCore.Qt.ConnectionType.UniqueConnection,
    # )
    conntype = QtCore.Qt.ConnectionType.QueuedConnection

    new_signal, new_slot = signal_slot

    try:
        slot_done_fn: Callable = slot_done(new_slot, new_signal)
        # MyPy/PyQt6-Stubs misses connect(type=param)

        new_signal.connect(slot_done_fn, type=conntype)  # type: ignore [attr-defined] # noqa: F821
    except Exception as error:  # pragma: no cover
        LOGGER.warning(  # pragma: no cover
            "ERROR Connecting: %s %s %s %s - %s error:%s",
            sender,
            signal,
            receiver,
            slot,
            error,
            conntype,
        )
        return None  # pragma: no cover

    signal_slot = new_signal, slot_done_fn
    return signal_slot


def disconnect(
    sender: "QtWidgets.QWidget", signal: str, receiver: "QtCore.QObject", slot: str
) -> Optional[Tuple["QtCore.pyqtSignal", Callable]]:
    """Disconnect signal from slot for QSA."""
    signal_slot = solve_connection(sender, signal, receiver, slot)
    if signal_slot:
        signal_, real_slot = signal_slot
        try:
            signal_.disconnect(real_slot)  # type: ignore [attr-defined] # noqa: F821
        except Exception:  # pragma: no cover
            LOGGER.trace(
                "Error disconnecting %r", (sender, signal, receiver, slot), exc_info=True
            )  # pragma: no cover

    return signal_slot


def solve_connection(
    sender: "QtWidgets.QWidget", signal: str, receiver: "QtCore.QObject", slot: str
) -> Optional[Tuple["QtCore.pyqtSignal", "Callable"]]:
    """Try hard to guess which is the correct way of connecting signal to slot. For QSA."""

    match = re.search(r"^(\w+)\.(\w+)(\(.*\))?", slot)
    if slot.endswith("()"):
        slot = slot[:-2]

    if hasattr(sender, "dateChanged"):
        if "valueChanged" in signal:
            signal = signal.replace("valueChanged", "dateChanged")

    elif hasattr(sender, "currentChanged"):
        if "CurrentChanged" in signal:
            signal = signal.replace("CurrentChanged", "currentChanged")

    remote_fn = None

    if slot.find(".") > -1:
        remote_fn = receiver
        for slot_ in slot.split("."):
            remote_fn = getattr(remote_fn, slot_, None)
            if remote_fn is None:
                break
    else:
        remote_fn = getattr(receiver, slot, None)

    sg_name = re.sub(r" *\(.*\)", "", signal)

    # search orig_signal
    original_signal = getattr(sender, sg_name, None)
    if not original_signal and hasattr(sender, "form"):
        original_signal = getattr(
            sender.form, sg_name, None  # type: ignore [attr-defined] # noqa: F821
        )

    if not original_signal:
        LOGGER.error(  # pragma: no cover
            "ERROR: No existe la señal %s para la clase %s", signal, sender.__class__.__name__
        )
        return None  # pragma: no cover

    if remote_fn is not None:
        proxy_slot = ProxySlot(remote_fn, receiver, slot)  # type: ignore [arg-type] # noqa F821
        proxyfn = proxy_slot.getProxyFn()
        return original_signal, proxyfn
    elif match:
        remote_obj = getattr(receiver, match.group(1), None)
        if remote_obj is None:
            raise AttributeError("Object %s not found on %s" % (remote_obj, str(receiver)))
        remote_fn = getattr(remote_obj, match.group(2), None)
        if remote_fn is None:
            raise AttributeError("Object %s not found on %s" % (remote_fn, remote_obj))
        return original_signal, remote_fn  # type: ignore [return-value] # noqa F723

    elif isinstance(receiver, QtCore.QObject):
        if isinstance(slot, str):
            original_slot = getattr(receiver, slot, None)
            if not original_slot:
                iface = getattr(receiver, "iface", None)
                if iface:
                    original_slot = getattr(iface, slot, None)
            if not original_slot:
                LOGGER.error(  # pragma: no cover
                    "Al realizar connect %s:%s -> %s:%s ; "
                    "el es QtCore.QObject pero no tiene slot",
                    sender,
                    signal,
                    receiver,
                    slot,
                )
                return None  # pragma: no cover
            return original_signal, original_slot



"""
Manage form sizes.
"""

from PyQt6 import QtCore  # type: ignore

from pineboolib.core import settings
from pineboolib import application


def save_geometry_form(name: str, geo: "QtCore.QSize") -> None:
    """
    Save the geometry of a window.

    @param name, window name.
    @param geo, QSize with window values.
    """

    if application.PROJECT.conn_manager is None:
        raise Exception("Project is not connected yet")

    name = "geo/%s/%s" % (application.PROJECT.conn_manager.mainConn().DBName(), name)
    settings.SETTINGS.set_value(name, geo)


def load_geometry_form(name: str) -> "QtCore.QSize":
    """
    Load the geometry of a window.

    @param name, window name
    @return QSize with the saved window geometry data.
    """
    if application.PROJECT.conn_manager is None:
        raise Exception("Project is not connected yet")

    name = "geo/%s/%s" % (application.PROJECT.conn_manager.mainConn().DBName(), name)
    value = settings.SETTINGS.value(name, None)
    if isinstance(value, dict):
        value = QtCore.QSize(int(value["width"]), int(value["height"]))
    return value



"""conn_dialog module."""

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # type: ignore[import] # pragma: no cover
    from pineboolib.loader.projectconfig import ProjectConfig  # pragma: no cover


def show_connection_dialog(app: "QtWidgets.QApplication") -> Optional["ProjectConfig"]:
    """Show the connection dialog, and configure the project accordingly."""
    from pineboolib.loader import dlgconnect

    connection_window = dlgconnect.DlgConnect()
    connection_window.load()
    connection_window.show()
    app.exec()  # FIXME: App should be started before this function
    connection_window.close()
    return connection_window.selected_project_config



"""Test_flutil module."""

import unittest

from pineboolib.fllegacy import flsettings


class TestSettings(unittest.TestCase):
    """TestSettings class."""

    def test_settings(self) -> None:
        """Test read functions."""

        setting = flsettings.FLSettings()
        setting.writeEntryList("test_uno", [""])
        setting.writeEntryList("test_uno", ["test_uno"])
        setting.writeEntryList("test_dos", [])
        setting.writeEntryList("test_dos", ["test_2_1", "test_2_2"])
        setting.writeEntry("test_tres", "")
        setting.writeEntry("test_tres", "test_tres")
        setting.writeEntry("test_cuatro", False)
        setting.writeEntry("test_cuatro", True)
        setting.writeEntry("test_cinco", 0)
        setting.writeEntry("test_cinco", 10)
        setting.writeEntry("test_double", 0.00)
        setting.writeEntry("test_double", 23.12)

        self.assertEqual(setting.readListEntry("test_dos"), ["test_2_1", "test_2_2"])
        self.assertEqual(setting.readListEntry("test_seis"), [])
        self.assertEqual(setting.readListEntry("test_uno"), ["test_uno"])
        self.assertEqual(setting.readEntry("test_tres"), "test_tres")
        self.assertEqual(setting.readEntry("test_siete", "fallo"), "fallo")
        self.assertEqual(setting.readNumEntry("test_cinco", 12), 10)
        self.assertEqual(setting.readNumEntry("test_ocho", 14), 14)
        self.assertTrue(setting.readBoolEntry("test_cuatro", False))
        self.assertTrue(setting.readBoolEntry("test_nueve", True))
        self.assertTrue(setting.readDoubleEntry("test_double", 23.12))



# -*- coding: utf-8 -*-
"""
AQSettings Module.

This class provides utilities to access local configuration data.
"""

from pineboolib.fllegacy.flsettings import FLSettings


class AQSettings(FLSettings):
    """AQSettings Class."""

    pass



"""qhbuttongroup module."""

from PyQt6 import QtWidgets  # type: ignore[import]

from pineboolib.q3widgets import qbuttongroup


class QVButtonGroup(qbuttongroup.QButtonGroup):
    """QHButtonGroup class."""

    def __init__(self, *args):
        """Initialize."""

        super().__init__(*args)
        self.setLayout(QtWidgets.QVBoxLayout())



# -*- coding: utf-8 -*-
"""
Create and register sqlAlchemy models in QSA tree.

Para llamar a uno de estos se puede hacer desde cualquier script de la siguiente manera

from pineboolib.qsa import *

cli = Clientes() <- Este es un modelo disponible, nombre del mtd existente y comenzando en Mayuscula.

También es posible recargar estos modelos creados a raiz de los mtds. Se puede crear por ejemplo
(tempdata)/cache/nombre_de_bd/models/Clientes_model.py. Esta librería sobrecargará en el arbol qsa la previa por defecto.

Un ejemplo sería:

    from pineboolib.qsa import *
    from sqlalchemy.orm import reconstructor

    class Clientes(Clientes):
        @reconstructor

        def init(self):
            print("Inicializado", self.nombre)


        def saluda(self):
            print("Hola", self.nombre)


Ejemplo de uso:
    from pineboolib.qsa import *

    session = aqApp.session()

    for instance in session.query(Clientes).order_by(Clientes.codcliente):
        instance.saluda()

"""
from pineboolib.application.utils import path
from pineboolib.application import load_script, qsadictmodules
from pineboolib.application.database.orm import basemodel
from pineboolib import logging, application
from pineboolib.application.parsers.parser_mtd import pnmtdparser
import sqlalchemy  # type: ignore [import]


from typing import Any, List, Dict, cast, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata  # pragma: no cover

LOGGER = logging.get_logger(__name__)
PROCESSED: List[str] = []
PENDING_RELATIONS = []


def register_metadata_as_model(metadata: "pntablemetadata.PNTableMetaData") -> bool:
    """Register a mtd as model."""

    name_ = metadata.name()

    if "%s_model" % name_ in PROCESSED:
        LOGGER.warning("Overwriting %s model" % name_)

    path_ = pnmtdparser.mtd_parse(metadata)
    return save_model(path_, name_)


def save_model(path_, name: str) -> bool:
    """Save model."""

    model_class = load_script.load_model(name, path_)

    if model_class is not None:
        # event.listen(model_class, "load", model_class._constructor_init)
        qsadictmodules.QSADictModules.set_qsa_tree(
            "%s_orm" % name, cast(basemodel.BaseModel, model_class)
        )
        sqlalchemy.event.listen(
            model_class,
            "load",
            model_class._constructor_init,  # type: ignore [attr-defined] # noqa: F821
        )

        for field in model_class.legacy_metadata[  # type: ignore [attr-defined] # noqa: F821
            "fields"
        ]:
            obj = getattr(model_class, field["name"], None)
            if obj is not None:
                sqlalchemy.event.listen(
                    obj,
                    "set",
                    model_class._changes_slot,  # type: ignore [attr-defined] # noqa: F821
                )

        if name not in PROCESSED:
            PROCESSED.append(name)

        func_ = getattr(model_class, "pending_relationships", None)
        if func_:
            PENDING_RELATIONS.append(func_)

        for pending_relation in PENDING_RELATIONS:
            if pending_relation():
                PENDING_RELATIONS.remove(pending_relation)

                LOGGER.info(
                    "Relaciones pendientes de %s , generadas al cargar %s"
                    % (pending_relation.__module__, name)
                )
                del pending_relation

        return True

    return False


def load_models() -> None:
    """Load all sqlAlchemy models."""
    # print("LOADING MODELS!!!")
    db_admin = application.PROJECT.db_admin_mode

    if application.PROJECT.conn_manager is None:
        raise Exception("Project is not connected yet")

    models_: Dict[str, Any] = {}
    views_: Dict[str, Any] = {}
    for action_name, action in application.PROJECT.actions.items():
        class_orm = action._class_orm

        if class_orm:
            if class_orm in PROCESSED:
                continue

            path_class_orm = path._path(class_orm, False)
            if not path_class_orm:
                LOGGER.warning(
                    "Se ha especificado un model (%s) en el action %s, pero el fichero no existe",
                    class_orm,
                    action_name,
                )
                continue

            models_[class_orm] = path_class_orm
            # print("***", class_orm)
            PROCESSED.append(class_orm)

    for key, file_ in application.PROJECT.files.items():
        if file_.filename.endswith("_model.py"):
            name = key[:-3]
            if name.endswith(".mtd_model"):
                name = "%s_model" % name[:-10]

            if name in PROCESSED:
                continue
            else:
                PROCESSED.append(name)
                models_[name] = file_.path()

    for key, data in models_.items():
        name = key[0:-6]
        save_model(data, name)
        metadata = application.PROJECT.conn_manager.manager().metadata(name, True)
        if metadata is not None:
            if metadata.isQuery():
                views_[name] = data
            else:
                application.PROJECT.conn_manager.manager().metadata(name)
    # views las últimas...
    for key, data in views_.items():
        save_model(data, key)
        if db_admin:
            application.PROJECT.conn_manager.manager().createTable(key)

    del models_
    del views_



# -*- coding: utf-8 -*-
"""Manage a parameters from a Query instance."""

from typing import Any


class PNParameterQuery(object):
    """Class that manages each parameter of a query."""

    name_: str
    alias_: str
    type_: int
    value_: Any

    def __init__(self, name: str, alias: str, type_: int) -> None:
        """Specify the name alias and type."""
        self.name_ = name
        self.alias_ = alias
        self.type_ = type_
        self.value_ = None

    def name(self) -> str:
        """Return parameter name."""

        return self.name_

    def alias(self) -> str:
        """Return parameter alias."""
        return self.alias_

    def type(self) -> int:
        """Return parameter type."""
        return self.type_

    def value(self) -> Any:
        """Return the value contained in the parameter."""

        return self.value_

    def setValue(self, value: Any) -> None:
        """Set a value for the parameter."""
        self.value_ = value



# -*- coding: utf-8 -*-
"""
PNBoolFlagState package.

Save cursor states.
"""
from typing import List, Optional


class PNBoolFlagState(object):
    """PNBollFlagState Class."""

    modifier_: str
    prev_value_: bool


class PNBoolFlagStateList(object):
    """PNBoolFlagStateList Class."""

    data_list_: List["PNBoolFlagState"]

    def __init__(self):
        """Initialize the list."""

        self.data_list_ = []

    def append(self, data: "PNBoolFlagState") -> None:
        """
        Add a state to the list.

        @param data. Flag state.
        """

        self.data_list_.append(data)

    def pushOnTop(self, data: "PNBoolFlagState") -> None:
        """
        Add a state to the list first.

        @param data. Flag state.
        """

        self.data_list_.insert(0, data)

    def erase(self, data: "PNBoolFlagState") -> None:
        """
        Erase a state to the list first.

        @param data. Flag state.
        """
        self.data_list_.remove(data)

    def find(self, data: str) -> Optional["PNBoolFlagState"]:
        """
        Search for a state in the list from a value.

        @param data. Value to search.
        @return Flag state.
        """

        for child in self.data_list_:
            if child.modifier_ == data:
                return child

        return None

    def current(self) -> Optional["PNBoolFlagState"]:
        """
        Return the first state of the list.

        @return Flag state.
        """

        return None if not self.data_list_ else self.data_list_[0]



"""Test_pnsqldrivers module."""

import unittest
from pineboolib import application

from pineboolib.loader.main import init_testing


class TestPNSqlDriversManager(unittest.TestCase):
    """TestPNSqlDriversManager Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_full(self) -> None:
        """Test full."""

        mng_ = application.PROJECT.conn_manager

        self.assertEqual(mng_._drivers_sql_manager.defaultDriverName(), "FLsqlite")
        self.assertEqual(mng_._drivers_sql_manager.driverName(), "FLsqlite")
        # self.assertTrue(
        #    conn_._driver_sql.isDesktopFile(
        #        conn_._driver_sql.nameToAlias(conn_._driver_sql.driverName())
        #    )
        # )
        self.assertEqual(
            mng_._drivers_sql_manager.port(
                mng_._drivers_sql_manager.nameToAlias(mng_._drivers_sql_manager.driverName())
            ),
            "0",
        )
        self.assertEqual(mng_._drivers_sql_manager.aliasToName(""), "FLsqlite")
        self.assertEqual(mng_._drivers_sql_manager.aliasToName(), "FLsqlite")



"""
Module for garbage collector checks.
"""
from typing import Any, Callable, List

from pineboolib.core.utils import logging
from pineboolib.core import DISABLE_CHECK_MEMORY_LEAKS

import weakref
import threading

import gc

LOGGER = logging.get_logger(__name__)


def check_delete(obj_: Any, name: str, force: bool = False) -> bool:
    """Delete a object."""

    if not DISABLE_CHECK_MEMORY_LEAKS or force:
        check_gc_referrers(obj_.__class__.__name__, weakref.ref(obj_), name)
    return True


def check_gc_referrers(typename: Any, w_obj: Callable, name: str) -> None:
    """
    Check if any variable is getting out of control.

    Great for checking and tracing memory leaks.
    """

    def checkfn() -> None:
        # time.sleep(2)
        list_: List[str] = []
        try:
            if w_obj is not None:
                gc.collect()
                obj = w_obj()
                if not obj:
                    return
                # TODO: Si ves el mensaje a continuación significa que "algo" ha dejado
                # ..... alguna referencia a un formulario (o similar) que impide que se destruya
                # ..... cuando se deja de usar. Causando que los connects no se destruyan tampoco
                # ..... y que se llamen referenciando al código antiguo y fallando.
                for ref in gc.get_referrers(obj):
                    if "<frame" in str(repr(ref)):
                        continue

                    elif isinstance(ref, dict):
                        for key, value in ref.items():
                            if value is obj:
                                list_.append(
                                    "(%s.%s -> %s (%s)" % (ref.__class__.__name__, key, name, ref)
                                )
                        # print(" - dict:", repr(x), gc.get_referrers(ref))
                    else:
                        list_.append(
                            "(%s) %s.%s -> %s (%s)"
                            % (ref.__class__.__name__, type(ref), str(repr(ref)), name, ref)
                        )
                        # print(" - obj:", repr(ref), [x for x in dir(ref) if getattr(ref, x) is obj])
                if list_:
                    LOGGER.warning("HINT: Objetos referenciando %r::%r (%r) :", typename, obj, name)
                    for item in list_:
                        LOGGER.warning(item)

        except Exception as error:  # noqa : F841
            LOGGER.warning("Error cleaning %r::%r (%r) :", typename, obj, name)

    threading.Thread(target=checkfn).start()



"""
test_pnunpacker Module.
"""

from pineboolib.application.packager.tests import fixture_path

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestPNUnpaker(unittest.TestCase):
    """TestUnpacker Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_pnunpacker(self) -> None:
        """Test eneboopkgs load."""
        from pineboolib.qsa import qsa
        from pineboolib.fllegacy import systype
        import os

        qsa_sys = systype.SysType()
        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa_sys.loadModules(path, False))
        qsa_sys.registerUpdate(path)

        qry = qsa.FLSqlQuery()
        qry.setTablesList("flfiles")
        qry.setSelect("count(nombre)")
        qry.setFrom("flfiles")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(qry.value(0), 148)

        qry_2 = qsa.FLSqlQuery()
        qry_2.setTablesList("flfiles")
        qry_2.setSelect("nombre")
        qry_2.setFrom("flfiles")
        qry_2.setWhere("nombre='impuestos.py'")
        self.assertTrue(qry_2.exec_())
        self.assertTrue(qry_2.first())
        self.assertEqual(qry_2.value(0), "impuestos.py")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Fltimeedit module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtCore  # type: ignore
from typing import Union, List


class FLTimeEdit(QtWidgets.QTimeEdit):
    """FLTimeEdit class."""

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        """Inicialize."""

        super().__init__(parent)
        self.setDisplayFormat("hh:mm:ss")
        self.setMinimumWidth(90)
        self.setTime(
            "00:00:00"
        )  # Normalmente deberia inicar a 00:00:00, pero inicia a 23:00:00 BUG??
        # self.setMaximumWidth(90)

    def setTime(self, value: Union[str, QtCore.QTime]) -> None:  # type: ignore
        """Set the time in the control."""

        if isinstance(value, str):
            list_: List[str] = value.split(":")
            value = QtCore.QTime(int(list_[0]), int(list_[1]), int(list_[2]))

        super().setTime(value)



"""Fluintvalidator module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtGui  # type: ignore[import]
from typing import Tuple, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # type: ignore[import] # noqa: F401 # pragma: no cover


class FLUIntValidator(QtGui.QIntValidator):
    """FLUItValidator class."""

    _formatting: bool

    def __init__(self, minimum: int, maximum: int, parent: Optional["QtWidgets.QWidget"]) -> None:
        """Inicialize."""

        super().__init__(minimum, maximum, parent)

        self._formatting = False

    def validate(
        self, input_: Optional[str], pos_cursor: int
    ) -> Tuple["QtGui.QValidator.State", str, int]:
        """Valiate a Value."""

        if not input_ or self._formatting:
            return (self.State.Acceptable, input_, pos_cursor)  # type: ignore [return-value]

        i_v = QtGui.QIntValidator(0, 1000000000, self)
        state = i_v.validate(input_, pos_cursor)

        ret_0 = self.State.Invalid if state[0] is self.State.Intermediate else state[0]
        ret_1 = state[1]
        ret_2 = state[2]

        return (ret_0, ret_1, ret_2)



"""
Tests for utils.version.VersionNumber.
"""
import unittest
from pineboolib.core.utils.version import VersionNumber


class TestVersionNumber(unittest.TestCase):
    """Test VersionNumber class."""

    def test_create(self) -> None:
        """Test create version."""
        version = VersionNumber("1.0.1")
        self.assertEqual(version.is_null, False)
        self.assertEqual(version.raw_text, "1.0.1")

    def test_normalize_complex(self) -> None:
        """Test normalize_complex function."""
        norm_cplx = VersionNumber.normalize_complex
        self.assertEqual(norm_cplx("1.0.0"), [(1, ""), (0, ""), (0, "")])
        self.assertEqual(norm_cplx("1b.0.0-ubuntu0"), [(1, "b"), (0, ""), (0, "-ubuntu0")])

    def test_compare(self) -> None:
        """Test compare versions."""

        self.assertEqual(VersionNumber("1.0.1") == VersionNumber("1.0.1 "), True)
        self.assertEqual(VersionNumber("1.0.1") > VersionNumber("1.0.1 "), False)
        self.assertEqual(VersionNumber("1.0b.1") < VersionNumber("1.0a.1 "), False)
        self.assertEqual(VersionNumber("1.0b.1") != VersionNumber("1.0a.1"), True)
        self.assertEqual(VersionNumber("1.0.1") >= VersionNumber("1.0.1"), True)
        self.assertEqual(VersionNumber("1.0.1") <= VersionNumber("1.0.1"), True)
        self.assertEqual(VersionNumber("1.0.1") > VersionNumber("1.0.1"), False)
        self.assertEqual(VersionNumber("1.0.1") < VersionNumber("1.0.1"), False)

    def test_repr_str(self) -> None:
        """Test repr and str methods."""
        self.assertEqual(str(VersionNumber("v1.0a.1 ")), "1.0a.1")
        self.assertEqual(str(VersionNumber(None)), "")
        self.assertEqual(repr(VersionNumber("v1.0a.1 ")), "<VersionNumber '1.0a.1'>")
        self.assertEqual(repr(VersionNumber(None)), "<VersionNumber None>")
